# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Encoders for text data.

* TextEncoder: base class
* ByteTextEncoder: for ascii text
* TokenTextEncoder: with user-supplied vocabulary file
* SubwordTextEncoder: invertible
* BertEncoder: for compatible tokenizers with original bert
"""

import collections
import itertools
import math
import re
import tempfile
import time
import unicodedata

from absl import logging
import numpy as np
import six
import tensorflow as tf
from trax.data import tokenizer

# Reserved tokens for things like padding and EOS symbols.
PAD = "<pad>"
EOS = "<EOS>"
RESERVED_TOKENS = [PAD, EOS]
NUM_RESERVED_TOKENS = len(RESERVED_TOKENS)
PAD_ID = RESERVED_TOKENS.index(PAD)  # Normally 0
EOS_ID = RESERVED_TOKENS.index(EOS)  # Normally 1
RESERVED_TOKENS_BYTES = [bytes(PAD, "ascii"), bytes(EOS, "ascii")]

# Regular expression for unescaping token strings.
# '\u' is converted to '_'
# '\\' is converted to '\'
# '\213;' is converted to unichr(213)
_UNESCAPE_REGEX = re.compile(r"\\u|\\\\|\\([0-9]+);")
_ESCAPE_CHARS = set(u"\\_u;0123456789")


# Unicode utility functions that work with Python 2 and 3
def native_to_unicode(s):
  if is_unicode(s):
    return s
  try:
    return to_unicode(s)
  except UnicodeDecodeError:
    res = to_unicode(s, ignore_errors=True)
    logging.info("Ignoring Unicode error, outputting: %s", res)
    return res


def is_unicode(s):
  return isinstance(s, six.text_type)


def to_unicode(s, ignore_errors=False):
  if is_unicode(s):
    return s
  error_mode = "ignore" if ignore_errors else "strict"
  return s.decode("utf-8", errors=error_mode)


def to_unicode_ignore_errors(s):
  return to_unicode(s, ignore_errors=True)


def to_unicode_utf8(s):
  return s.decode("utf-8")


def strip_ids(ids, ids_to_strip):
  """Strip ids_to_strip from the end IDs."""
  ids = list(ids)
  while ids and ids[-1] in ids_to_strip:
    ids.pop()
  return ids


class TextEncoder:
  """Base class for converting from ints to/from human readable strings."""

  def __init__(self, num_reserved_ids=NUM_RESERVED_TOKENS):
    self._num_reserved_ids = num_reserved_ids

  @property
  def num_reserved_ids(self):
    return self._num_reserved_ids

  def encode(self, s):
    """Transform a human-readable string into a sequence of int IDs.

    The IDs should be in the range [num_reserved_ids, vocab_size). IDs [0,
    num_reserved_ids) are reserved.

    EOS is not appended.

    Args:
      s: human-readable string to be converted.

    Returns:
      ids: list of integers
    """
    return [int(w) + self._num_reserved_ids for w in s.split()]

  def decode(self, ids, strip_extraneous=False):
    """Transform a sequence of int IDs into a human-readable string.

    EOS is not expected in IDs.

    Args:
      ids: list of integers to be converted.
      strip_extraneous: bool, whether to strip off extraneous tokens (EOS and
        PAD).

    Returns:
      s: human-readable string.
    """
    if strip_extraneous:
      ids = strip_ids(ids, list(range(self._num_reserved_ids or 0)))
    return " ".join(self.decode_list(ids))

  def decode_list(self, ids):
    """Transform a sequence of int IDs into a their string versions.

    This method supports transforming individual input/output IDs to their
    string versions so that sequence to/from text conversions can be visualized
    in a human readable format.

    Args:
      ids: list of integers to be converted.

    Returns:
      strs: list of human-readable string.
    """
    decoded_ids = []
    for id_ in ids:
      if 0 <= id_ < self._num_reserved_ids:
        decoded_ids.append(RESERVED_TOKENS[int(id_)])
      else:
        decoded_ids.append(id_ - self._num_reserved_ids)
    return [str(d) for d in decoded_ids]

  @property
  def vocab_size(self):
    raise NotImplementedError()


class ByteTextEncoder(TextEncoder):
  """Encodes each byte to an id. For 8-bit strings only."""

  def encode(self, s):
    numres = self._num_reserved_ids
    # Python3: explicitly convert to UTF-8
    return [c + numres for c in s.encode("utf-8")]

  def decode(self, ids, strip_extraneous=False):
    if strip_extraneous:
      ids = strip_ids(ids, list(range(self._num_reserved_ids or 0)))
    numres = self._num_reserved_ids
    decoded_ids = []
    int2byte = six.int2byte
    for id_ in ids:
      if 0 <= id_ < numres:
        decoded_ids.append(RESERVED_TOKENS_BYTES[int(id_)])
      else:
        decoded_ids.append(int2byte(id_ - numres))
    # Python3: join byte arrays and then decode string
    return b"".join(decoded_ids).decode("utf-8", "replace")

  def decode_list(self, ids):
    numres = self._num_reserved_ids
    decoded_ids = []
    int2byte = six.int2byte
    for id_ in ids:
      if 0 <= id_ < numres:
        decoded_ids.append(RESERVED_TOKENS_BYTES[int(id_)])
      else:
        decoded_ids.append(int2byte(id_ - numres))
    # Python3: join byte arrays and then decode string
    return decoded_ids

  @property
  def vocab_size(self):
    return 2**8 + self._num_reserved_ids


class ClassLabelEncoder(TextEncoder):
  """Encoder for class labels."""

  def __init__(self, class_labels=None, class_labels_fname=None):
    super(ClassLabelEncoder, self).__init__(num_reserved_ids=0)

    if class_labels_fname:
      with tf.io.gfile.GFile(class_labels_fname) as f:
        class_labels = [label.strip() for label in f.readlines()]

    assert class_labels
    self._class_labels = class_labels

  def encode(self, s):
    label_str = s
    return self._class_labels.index(label_str)

  def decode(self, ids, strip_extraneous=False):
    del strip_extraneous
    label_id = ids
    if isinstance(label_id, list):
      assert len(label_id) == 1
      label_id, = label_id
    if isinstance(label_id, np.ndarray):
      label_id = np.squeeze(label_id)
    return self._class_labels[label_id]

  def decode_list(self, ids):
    return [self._class_labels[i] for i in ids]

  @property
  def vocab_size(self):
    return len(self._class_labels)


class OneHotClassLabelEncoder(ClassLabelEncoder):
  """One-hot encoder for class labels."""

  def encode(self, label_str, on_value=1, off_value=0):  # pylint: disable=arguments-differ
    e = np.full(self.vocab_size, off_value, dtype=np.int32)
    e[self._class_labels.index(label_str)] = on_value
    return e.tolist()

  def decode(self, ids, strip_extraneous=False):
    del strip_extraneous
    label_id = ids
    if isinstance(label_id, np.ndarray):
      label_id = np.squeeze(label_id).astype(np.int8).tolist()
    assert isinstance(label_id, list)
    assert len(label_id) == self.vocab_size
    return self._class_labels[label_id.index(1)]

  @property
  def vocab_size(self):
    return len(self._class_labels)


class TokenTextEncoder(TextEncoder):
  """Encoder based on a user-supplied vocabulary (file or list)."""

  def __init__(self,
               vocab_filename,
               reverse=False,
               vocab_list=None,
               replace_oov=None,
               num_reserved_ids=NUM_RESERVED_TOKENS):
    """Initialize from a file or list, one token per line.

    Handling of reserved tokens works as follows:
    - When initializing from a list, we add reserved tokens to the vocab.
    - When initializing from a file, we do not add reserved tokens to the vocab.
    - When saving vocab files, we save reserved tokens to the file.

    Args:
      vocab_filename: If not None, the full filename to read vocab from. If this
        is not None, then vocab_list should be None.
      reverse: Boolean indicating if tokens should be reversed during encoding
        and decoding.
      vocab_list: If not None, a list of elements of the vocabulary. If this is
        not None, then vocab_filename should be None.
      replace_oov: If not None, every out-of-vocabulary token seen when encoding
        will be replaced by this string (which must be in vocab).
      num_reserved_ids: Number of IDs to save for reserved tokens like <EOS>.
    """
    super(TokenTextEncoder, self).__init__(num_reserved_ids=num_reserved_ids)
    self._reverse = reverse
    self._replace_oov = replace_oov
    if vocab_filename:
      self._init_vocab_from_file(vocab_filename)
    else:
      assert vocab_list is not None
      self._init_vocab_from_list(vocab_list)

  def encode(self, s):
    """Converts a space-separated string of tokens to a list of ids."""
    sentence = s
    tokens = sentence.strip().split()
    if self._replace_oov is not None:
      tokens = [
          t if t in self._token_to_id else self._replace_oov for t in tokens
      ]
    ret = [self._token_to_id[tok] for tok in tokens]
    return ret[::-1] if self._reverse else ret

  def decode(self, ids, strip_extraneous=False):
    return " ".join(self.decode_list(ids))

  def decode_list(self, ids):
    seq = reversed(ids) if self._reverse else ids
    return [self._safe_id_to_token(i) for i in seq]

  @property
  def vocab_size(self):
    return len(self._id_to_token)

  def _safe_id_to_token(self, idx):
    return self._id_to_token.get(idx, "ID_%d" % idx)

  def _init_vocab_from_file(self, filename):
    """Load vocab from a file.

    Args:
      filename: The file to load vocabulary from.
    """
    with tf.io.gfile.GFile(filename) as f:
      tokens = [token.strip() for token in f.readlines()]

    def token_gen():
      for token in tokens:
        yield token

    self._init_vocab(token_gen(), add_reserved_tokens=False)

  def _init_vocab_from_list(self, vocab_list):
    """Initialize tokens from a list of tokens.

    It is ok if reserved tokens appear in the vocab list. They will be
    removed. The set of tokens in vocab_list should be unique.

    Args:
      vocab_list: A list of tokens.
    """

    def token_gen():
      for token in vocab_list:
        if token not in RESERVED_TOKENS:
          yield token

    self._init_vocab(token_gen())

  def _init_vocab(self, token_generator, add_reserved_tokens=True):
    """Initialize vocabulary with tokens from token_generator."""

    self._id_to_token = {}
    non_reserved_start_index = 0

    if add_reserved_tokens:
      self._id_to_token.update(enumerate(RESERVED_TOKENS))
      non_reserved_start_index = len(RESERVED_TOKENS)

    self._id_to_token.update(
        enumerate(token_generator, start=non_reserved_start_index))

    # _token_to_id is the reverse of _id_to_token
    self._token_to_id = dict(
        (v, k) for k, v in six.iteritems(self._id_to_token))

  def store_to_file(self, filename):
    """Write vocab file to disk.

    Vocab files have one token per line. The file ends in a newline. Reserved
    tokens are written to the vocab file as well.

    Args:
      filename: Full path of the file to store the vocab to.
    """
    with tf.io.gfile.GFile(filename, "w") as f:
      for i in range(len(self._id_to_token)):
        f.write(self._id_to_token[i] + "\n")


def _escape_token(token, alphabet):
  """Escape away underscores and OOV characters and append '_'.

  This allows the token to be expressed as the concatenation of a list
  of subtokens from the vocabulary. The underscore acts as a sentinel
  which allows us to invertibly concatenate multiple such lists.

  Args:
    token: A unicode string to be escaped.
    alphabet: A set of all characters in the vocabulary's alphabet.

  Returns:
    escaped_token: An escaped unicode string.

  Raises:
    ValueError: If the provided token is not unicode.
  """
  if not isinstance(token, six.text_type):
    raise ValueError("Expected string type for token, got %s" % type(token))

  token = token.replace(u"\\", u"\\\\").replace(u"_", u"\\u")
  ret = [c if c in alphabet and c != u"\n" else r"\%d;" % ord(c) for c in token]
  return u"".join(ret) + "_"


def _unescape_token(escaped_token):
  """Inverse of _escape_token().

  Args:
    escaped_token: a unicode string

  Returns:
    token: a unicode string
  """

  def match(m):
    if m.group(1) is None:
      return u"_" if m.group(0) == u"\\u" else u"\\"

    try:
      return six.unichr(int(m.group(1)))
    except (ValueError, OverflowError) as _:
      return u"\u3013"  # Unicode for undefined character.

  trimmed = escaped_token[:-1] if escaped_token.endswith("_") else escaped_token
  return _UNESCAPE_REGEX.sub(match, trimmed)


class SubwordTextEncoder(TextEncoder):
  """Class for invertibly encoding text using a limited vocabulary.

  Invertibly encodes a native string as a sequence of subtokens from a limited
  vocabulary.

  A SubwordTextEncoder is built from a corpus (so it is tailored to the text in
  the corpus), and stored to a file. See text_encoder_build_subword.py.

  It can then be loaded and used to encode/decode any text.

  Encoding has four phases:

  1. Tokenize into a list of tokens.  Each token is a unicode string of either
     all alphanumeric characters or all non-alphanumeric characters.  We drop
     tokens consisting of a single space that are between two alphanumeric
     tokens.

  2. Escape each token.  This escapes away special and out-of-vocabulary
     characters, and makes sure that each token ends with an underscore, and
     has no other underscores.

  3. Represent each escaped token as a the concatenation of a list of subtokens
     from the limited vocabulary.  Subtoken selection is done greedily from
     beginning to end.  That is, we construct the list in order, always picking
     the longest subtoken in our vocabulary that matches a prefix of the
     remaining portion of the encoded token.

  4. Concatenate these lists.  This concatenation is invertible due to the
     fact that the trailing underscores indicate when one list is finished.

  """

  def __init__(self, filename=None):
    """Initialize and read from a file, if provided.

    Args:
      filename: filename from which to read vocab. If None, do not load a vocab
    """
    self._alphabet = set()
    self.filename = filename
    if filename is not None:
      self._load_from_file(filename)
    super(SubwordTextEncoder, self).__init__()

  def encode(self, s):
    """Converts a native string to a list of subtoken IDs.

    Args:
      s: a native string.

    Returns:
      a list of integers in the range [0, vocab_size)
    """
    return self._tokens_to_subtoken_ids(tokenizer.encode(native_to_unicode(s)))

  def encode_without_tokenizing(self, token_text):
    """Converts string to list of subtoken IDs without calling tokenizer.

    This treats `token_text` as a single token and directly converts it
    to subtoken IDs. This may be useful when the default tokenizer doesn't
    do what we want (e.g., when encoding text with tokens composed of lots of
    nonalphanumeric characters). It is then up to the caller to make sure that
    raw text is consistently converted into tokens. Only use this if you are
    sure that `encode` doesn't suit your needs.

    Args:
      token_text: A native string representation of a single token.

    Returns:
      A list of subword token IDs; i.e., integers in the range [0, vocab_size).
    """
    return self._tokens_to_subtoken_ids([native_to_unicode(token_text)])

  def decode(self, ids, strip_extraneous=False):
    """Converts a sequence of subtoken IDs to a native string.

    Args:
      ids: a list of integers in the range [0, vocab_size)
      strip_extraneous: bool, whether to strip off extraneous tokens (EOS and
        PAD).

    Returns:
      a native string
    """
    if strip_extraneous:
      ids = strip_ids(ids, list(range(self._num_reserved_ids or 0)))
    return tokenizer.decode(self._subtoken_ids_to_tokens(ids))

  def decode_list(self, ids):
    return [self._subtoken_id_to_subtoken_string(s) for s in ids]

  @property
  def vocab_size(self):
    """The subtoken vocabulary size."""
    return len(self._all_subtoken_strings)

  def _tokens_to_subtoken_ids(self, tokens):
    """Converts a list of tokens to a list of subtoken IDs.

    Args:
      tokens: a list of strings.

    Returns:
      a list of integers in the range [0, vocab_size)
    """
    ret = []
    for token in tokens:
      ret.extend(self._token_to_subtoken_ids(token))
    return ret

  def _token_to_subtoken_ids(self, token):
    """Converts token to a list of subtoken IDs.

    Args:
      token: a string.

    Returns:
      a list of integers in the range [0, vocab_size)
    """
    cache_location = hash(token) % self._cache_size
    cache_key, cache_value = self._cache[cache_location]
    if cache_key == token:
      return cache_value
    ret = self._escaped_token_to_subtoken_ids(
        _escape_token(token, self._alphabet))
    self._cache[cache_location] = (token, ret)
    return ret

  def _subtoken_ids_to_tokens(self, subtokens):
    """Converts a list of subtoken IDs to a list of tokens.

    Args:
      subtokens: a list of integers in the range [0, vocab_size)

    Returns:
      a list of strings.
    """
    concatenated = "".join(
        [self._subtoken_id_to_subtoken_string(s) for s in subtokens])
    split = concatenated.split("_")
    ret = []
    for t in split:
      if t:
        unescaped = _unescape_token(t + "_")
        if unescaped:
          ret.append(unescaped)
    return ret

  def _subtoken_id_to_subtoken_string(self, subtoken):
    """Converts a subtoken integer ID to a subtoken string."""
    if 0 <= subtoken < self.vocab_size:
      return self._all_subtoken_strings[subtoken]
    return u""

  def _escaped_token_to_subtoken_strings(self, escaped_token):
    """Converts an escaped token string to a list of subtoken strings.

    Args:
      escaped_token: An escaped token as a unicode string.

    Returns:
      A list of subtokens as unicode strings.
    """
    # NOTE: This algorithm is greedy; it won't necessarily produce the "best"
    # list of subtokens.
    ret = []
    start = 0
    token_len = len(escaped_token)
    while start < token_len:
      for end in range(
          min(token_len, start + self._max_subtoken_len), start, -1):
        subtoken = escaped_token[start:end]
        if subtoken in self._subtoken_string_to_id:
          ret.append(subtoken)
          start = end
          break

      else:  # Did not break
        # If there is no possible encoding of the escaped token then one of the
        # characters in the token is not in the alphabet. This should be
        # impossible and would be indicative of a bug.
        assert False, "Token substring not found in subtoken vocabulary."

    return ret

  def _escaped_token_to_subtoken_ids(self, escaped_token):
    """Converts an escaped token string to a list of subtoken IDs.

    Args:
      escaped_token: An escaped token as a unicode string.

    Returns:
      A list of subtoken IDs as integers.
    """
    return [
        self._subtoken_string_to_id[subtoken]
        for subtoken in self._escaped_token_to_subtoken_strings(escaped_token)
    ]

  @classmethod
  def build_from_generator(cls,
                           generator,
                           target_size,
                           max_subtoken_length=None,
                           reserved_tokens=None):
    """Builds a SubwordTextEncoder from the generated text.

    Args:
      generator: yields text.
      target_size: int, approximate vocabulary size to create.
      max_subtoken_length: Maximum length of a subtoken. If this is not set,
        then the runtime and memory use of creating the vocab is quadratic in
        the length of the longest token. If this is set, then it is instead
        O(max_subtoken_length * length of longest token).
      reserved_tokens: List of reserved tokens. The global variable
        `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
        argument is `None`, it will use `RESERVED_TOKENS`.

    Returns:
      SubwordTextEncoder with `vocab_size` approximately `target_size`.
    """
    token_counts = collections.defaultdict(int)
    for item in generator:
      for tok in tokenizer.encode(native_to_unicode(item)):
        token_counts[tok] += 1
    encoder = cls.build_to_target_size(
        target_size,
        token_counts,
        1,
        1e3,
        max_subtoken_length=max_subtoken_length,
        reserved_tokens=reserved_tokens)
    return encoder

  @classmethod
  def build_to_target_size(cls,
                           target_size,
                           token_counts,
                           min_val,
                           max_val,
                           max_subtoken_length=None,
                           reserved_tokens=None,
                           num_iterations=4):
    """Builds a SubwordTextEncoder that has `vocab_size` near `target_size`.

    Uses simple recursive binary search to find a minimum token count that most
    closely matches the `target_size`.

    Args:
      target_size: Desired vocab_size to approximate.
      token_counts: A dictionary of token counts, mapping string to int.
      min_val: An integer; lower bound for the minimum token count.
      max_val: An integer; upper bound for the minimum token count.
      max_subtoken_length: Maximum length of a subtoken. If this is not set,
        then the runtime and memory use of creating the vocab is quadratic in
        the length of the longest token. If this is set, then it is instead
        O(max_subtoken_length * length of longest token).
      reserved_tokens: List of reserved tokens. The global variable
        `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
        argument is `None`, it will use `RESERVED_TOKENS`.
      num_iterations: An integer; how many iterations of refinement.

    Returns:
      A SubwordTextEncoder instance.

    Raises:
      ValueError: If `min_val` is greater than `max_val`.
    """
    if min_val > max_val:
      raise ValueError("Lower bound for the minimum token count "
                       "is greater than the upper bound.")
    if target_size < 1:
      raise ValueError("Target size must be positive.")

    if reserved_tokens is None:
      reserved_tokens = RESERVED_TOKENS

    def bisect(min_val, max_val):
      """Bisection to find the right size."""
      present_count = (max_val + min_val) // 2
      logging.info("Trying min_count %d", present_count)
      subtokenizer = cls()
      subtokenizer.build_from_token_counts(
          token_counts,
          present_count,
          num_iterations,
          max_subtoken_length=max_subtoken_length,
          reserved_tokens=reserved_tokens)

      # Being within 1% of the target size is ok.
      is_ok = abs(subtokenizer.vocab_size - target_size) * 100 < target_size
      # If min_val == max_val, we can't do any better than this.
      if is_ok or min_val >= max_val or present_count < 2:
        return subtokenizer

      if subtokenizer.vocab_size > target_size:
        other_subtokenizer = bisect(present_count + 1, max_val)
      else:
        other_subtokenizer = bisect(min_val, present_count - 1)

      if other_subtokenizer is None:
        return subtokenizer

      if (abs(other_subtokenizer.vocab_size - target_size) <
          abs(subtokenizer.vocab_size - target_size)):
        return other_subtokenizer
      return subtokenizer

    return bisect(min_val, max_val)

  def build_from_token_counts(self,
                              token_counts,
                              min_count,
                              num_iterations=4,
                              reserved_tokens=None,
                              max_subtoken_length=None):
    """Train a SubwordTextEncoder based on a dictionary of word counts.

    Args:
      token_counts: a dictionary of Unicode strings to int.
      min_count: an integer - discard subtokens with lower counts.
      num_iterations: an integer.  how many iterations of refinement.
      reserved_tokens: List of reserved tokens. The global variable
        `RESERVED_TOKENS` must be a prefix of `reserved_tokens`. If this
        argument is `None`, it will use `RESERVED_TOKENS`.
      max_subtoken_length: Maximum length of a subtoken. If this is not set,
        then the runtime and memory use of creating the vocab is quadratic in
        the length of the longest token. If this is set, then it is instead
        O(max_subtoken_length * length of longest token).

    Raises:
      ValueError: if reserved is not 0 or len(RESERVED_TOKENS). In this case, it
        is not clear what the space is being reserved for, or when it will be
        filled in.
    """
    if reserved_tokens is None:
      reserved_tokens = RESERVED_TOKENS
    else:
      # There is not complete freedom in replacing RESERVED_TOKENS.
      for default, proposed in zip(RESERVED_TOKENS, reserved_tokens):
        if default != proposed:
          raise ValueError("RESERVED_TOKENS must be a prefix of "
                           "reserved_tokens.")

    # Initialize the alphabet. Note, this must include reserved tokens or it can
    # result in encoding failures.
    alphabet_tokens = itertools.chain(
        six.iterkeys(token_counts),
        [native_to_unicode(t) for t in reserved_tokens])

    self._init_alphabet_from_tokens(alphabet_tokens)

    # Bootstrap the initial list of subtokens with the characters from the
    # alphabet plus the escaping characters.
    self._init_subtokens_from_list(
        list(self._alphabet), reserved_tokens=reserved_tokens)

    # We build iteratively.  On each iteration, we segment all the words,
    # then count the resulting potential subtokens, keeping the ones
    # with high enough counts for our new vocabulary.
    if min_count < 1:
      min_count = 1
    for i in range(num_iterations):
      logging.info("Iteration %d", i)

      # Collect all substrings of the encoded token that break along current
      # subtoken boundaries.
      subtoken_counts = collections.defaultdict(int)
      for token, count in six.iteritems(token_counts):
        iter_start_time = time.time()
        escaped_token = _escape_token(token, self._alphabet)
        subtokens = self._escaped_token_to_subtoken_strings(escaped_token)
        start = 0
        for subtoken in subtokens:
          last_position = len(escaped_token) + 1
          if max_subtoken_length is not None:
            last_position = min(last_position, start + max_subtoken_length)

          for end in range(start + 1, last_position):
            new_subtoken = escaped_token[start:end]
            subtoken_counts[new_subtoken] += count
          start += len(subtoken)
        iter_time_secs = time.time() - iter_start_time
        if iter_time_secs > 0.1:
          logging.info(
              "Processing token [%s] took {%d} seconds, consider "
              "setting Text2TextProblem.max_subtoken_length to a "
              "smaller value.", token, iter_time_secs)

      # Array of sets of candidate subtoken strings, by length.
      len_to_subtoken_strings = []
      for subtoken_string, count in six.iteritems(subtoken_counts):
        lsub = len(subtoken_string)
        if count >= min_count:
          while len(len_to_subtoken_strings) <= lsub:
            len_to_subtoken_strings.append(set())
          len_to_subtoken_strings[lsub].add(subtoken_string)

      # Consider the candidates longest to shortest, so that if we accept
      # a longer subtoken string, we can decrement the counts of its prefixes.
      new_subtoken_strings = []
      for lsub in range(len(len_to_subtoken_strings) - 1, 0, -1):
        subtoken_strings = len_to_subtoken_strings[lsub]
        for subtoken_string in subtoken_strings:
          count = subtoken_counts[subtoken_string]
          if count >= min_count:
            # Exclude alphabet tokens here, as they must be included later,
            # explicitly, regardless of count.
            if subtoken_string not in self._alphabet:
              new_subtoken_strings.append((count, subtoken_string))
            for l in range(1, lsub):
              subtoken_counts[subtoken_string[:l]] -= count

      # Include the alphabet explicitly to guarantee all strings are encodable.
      new_subtoken_strings.extend(
          (subtoken_counts.get(a, 0), a) for a in self._alphabet)
      new_subtoken_strings.sort(reverse=True)

      # Reinitialize to the candidate vocabulary.
      new_subtoken_strings = [subtoken for _, subtoken in new_subtoken_strings]
      if reserved_tokens:
        escaped_reserved_tokens = [
            _escape_token(native_to_unicode(t), self._alphabet)
            for t in reserved_tokens
        ]
        new_subtoken_strings = escaped_reserved_tokens + new_subtoken_strings

      self._init_subtokens_from_list(new_subtoken_strings)
      logging.info("vocab_size = %d", self.vocab_size)

  @property
  def all_subtoken_strings(self):
    return tuple(self._all_subtoken_strings)

  def dump(self):
    """Debugging dump of the current subtoken vocabulary."""
    subtoken_strings = [
        (i, s) for s, i in six.iteritems(self._subtoken_string_to_id)
    ]
    print(u", ".join(
        u"{0} : '{1}'".format(i, s) for i, s in sorted(subtoken_strings)))

  def _init_subtokens_from_list(self, subtoken_strings, reserved_tokens=None):
    """Initialize token information from a list of subtoken strings.

    Args:
      subtoken_strings: a list of subtokens
      reserved_tokens: List of reserved tokens. We must have `reserved_tokens`
        as None or the empty list, or else the global variable `RESERVED_TOKENS`
        must be a prefix of `reserved_tokens`.

    Raises:
      ValueError: if reserved is not 0 or len(RESERVED_TOKENS). In this case, it
        is not clear what the space is being reserved for, or when it will be
        filled in.
    """
    if reserved_tokens is None:
      reserved_tokens = []

    if reserved_tokens:
      self._all_subtoken_strings = reserved_tokens + subtoken_strings
    else:
      self._all_subtoken_strings = subtoken_strings

    # we remember the maximum length of any subtoken to avoid having to
    # check arbitrarily long strings.
    self._max_subtoken_len = max([len(s) for s in subtoken_strings])
    self._subtoken_string_to_id = {
        s: i + len(reserved_tokens) for i, s in enumerate(subtoken_strings) if s
    }
    # Initialize the cache to empty.
    self._cache_size = 2**20
    self._cache = [(None, None)] * self._cache_size

  def _init_alphabet_from_tokens(self, tokens):
    """Initialize alphabet from an iterable of token or subtoken strings."""
    # Include all characters from all tokens in the alphabet to guarantee that
    # any token can be encoded. Additionally, include all escaping characters.
    self._alphabet = {c for token in tokens for c in token}  # pylint: disable=g-complex-comprehension
    self._alphabet |= _ESCAPE_CHARS

  def _load_from_file_object(self, f):
    """Load from a file object.

    Args:
      f: File object to load vocabulary from
    """
    subtoken_strings = []
    for line in f:
      s = line.rstrip()
      # Some vocab files wrap words in single quotes, but others don't
      if ((s.startswith("'") and s.endswith("'")) or
          (s.startswith("\"") and s.endswith("\""))):
        s = s[1:-1]
      subtoken_strings.append(native_to_unicode(s))
    self._init_subtokens_from_list(subtoken_strings)
    self._init_alphabet_from_tokens(subtoken_strings)

  def _load_from_file(self, filename):
    """Load from a vocab file."""
    if not tf.io.gfile.exists(filename):
      raise ValueError("File %s not found" % filename)
    with tf.io.gfile.GFile(filename) as f:
      self._load_from_file_object(f)

  def store_to_file(self, filename, add_single_quotes=True):
    with tf.io.gfile.GFile(filename, "w") as f:
      for subtoken_string in self._all_subtoken_strings:
        if add_single_quotes:
          f.write("'" + subtoken_string + "'\n")
        else:
          f.write(subtoken_string + "\n")


class ImageEncoder:
  """Encoder class for saving and loading images."""

  def __init__(self, num_reserved_ids=0, height=None, width=None, channels=3):
    assert num_reserved_ids == 0
    self._height = height
    self._width = width
    self._channels = channels

  @property
  def num_reserved_ids(self):
    return 0

  def encode(self, s):
    """Transform a string with a filename into a list of RGB integers.

    Args:
      s: path to the file with an image.

    Returns:
      ids: list of integers
    """
    try:
      import matplotlib.image as im  # pylint: disable=g-import-not-at-top
    except ImportError as e:
      logging.warning(
          "Reading an image requires matplotlib to be installed: %s", e)
      raise NotImplementedError("Image reading not implemented.")
    return im.imread(s)

  def decode(self, ids, strip_extraneous=False):
    """Transform a sequence of int IDs into an image file.

    Args:
      ids: list of integers to be converted.
      strip_extraneous: unused

    Returns:
      Path to the temporary file where the image was saved.

    Raises:
      ValueError: if the IDs are not of the appropriate size.
    """
    del strip_extraneous
    _, tmp_file_path = tempfile.mkstemp("_decode.png")
    if self._height is None or self._width is None:
      size = int(math.sqrt(len(ids) / self._channels))
      length = size * size * self._channels
    else:
      size = None
      length = self._height * self._width * self._channels
    if len(ids) != length:
      raise ValueError("Length of ids (%d) must be height (%d) x width (%d) x "
                       "channels (%d); %d != %d.\n Ids: %s" %
                       (len(ids), self._height, self._width, self._channels,
                        len(ids), length, " ".join([str(i) for i in ids])))
    with tf.Graph().as_default():
      raw = tf.constant(ids, dtype=tf.uint8)
      if size is None:
        img = tf.reshape(raw, [self._height, self._width, self._channels])
      else:
        img = tf.reshape(raw, [size, size, self._channels])
      png = tf.image.encode_png(img)
      op = tf.write_file(tmp_file_path, png)
      with tf.Session() as sess:
        sess.run(op)
    return tmp_file_path

  def decode_list(self, ids):
    """Transform a sequence of int IDs into an image file.

    Args:
      ids: list of integers to be converted.

    Returns:
      Singleton list: path to the temporary file where the image was saved.
    """
    return [self.decode(ids)]

  @property
  def vocab_size(self):
    return 256


class RealEncoder:
  """Encoder class for saving and loading float values."""

  def encode(self, s):
    """Transform a string (space separated float values) into a float array.

    Args:
      s: space separated float values.

    Returns:
      Array of float values.
    """
    return [float(w) for w in s.split()]

  def decode(self, ids, strip_extraneous=False):
    """Transform sequence of float values into string (float values).

    Args:
      ids: array of floats to be converted.
      strip_extraneous: unused

    Returns:
      String having space separated float values.

    Raises:
      ValueError: if the IDs are not of the appropriate size.
    """
    del strip_extraneous
    return " ".join([str(i) for i in ids])


class BertEncoder:
  """Encoder Class that is compatible with models trained in original BERT library."""

  def __init__(self, vocab_file, do_lower_case=True):
    self._vocab = self.load_vocab(vocab_file)
    self._inv_vocab = {v: k for k, v in self._vocab.items()}
    self._basic_tokenizer = BertBasicEncoder(do_lower_case=do_lower_case)
    self._wordpiece_tokenizer = BertWordpieceTokenizer(vocab=self._vocab)

  def load_vocab(self, vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    index = 0
    with tf.io.gfile.GFile(vocab_file, "r") as reader:
      while True:
        token = native_to_unicode(reader.readline())
        if not token:
          break
        token = token.strip()
        vocab[token] = index
        index += 1
    return vocab

  def encode(self, text):
    return self._convert_tokens_to_ids(self.tokenize(text))

  # Note: Because encoding by BertEncoder is not unique text decoded
  # from token ids is not unique.
  def decode(self, ids):
    """Returns a text that encoded would yield provided ids."""
    tokens = self._convert_ids_to_tokens(ids)
    if not tokens:
      return ""
    retarr = [tokens[0]]
    for token in tokens[1:]:
      if token.startswith("##"):
        retarr.append(token.lstrip("#"))
      else:
        retarr.append(" ")
        retarr.append(token)
    return "".join(retarr)

  @property
  def vocab_size(self):
    return len(self._vocab)

  def tokenize(self, text):
    split_tokens = []
    for token in self._basic_tokenizer.tokenize(text):
      for sub_token in self._wordpiece_tokenizer.tokenize(token):
        split_tokens.append(sub_token)

    return split_tokens

  def _convert_tokens_to_ids(self, tokens):
    return [self._vocab[token] for token in tokens]

  def _convert_ids_to_tokens(self, ids):
    return [self._inv_vocab[token_id] for token_id in ids]


class BertBasicEncoder:
  """Part of BertEncoder; tokenization (punctuation splitting, lower casing)."""

  def __init__(self, do_lower_case=True):
    """Constructs a BasicTokenizer.

    Args:
      do_lower_case: Whether to lower case the input.
    """
    self.do_lower_case = do_lower_case

  def tokenize(self, text):
    """Tokenizes a piece of text."""
    text = native_to_unicode(text)
    text = self._clean_text(text)

    text = self._tokenize_chinese_chars(text)

    orig_tokens = whitespace_tokenize(text)
    split_tokens = []
    for token in orig_tokens:
      if self.do_lower_case:
        token = token.lower()
        token = self._run_strip_accents(token)
      split_tokens.extend(self._run_split_on_punc(token))

    output_tokens = whitespace_tokenize(" ".join(split_tokens))
    return output_tokens

  def _run_strip_accents(self, text):
    """Strips accents from a piece of text."""
    text = unicodedata.normalize("NFD", text)
    output = []
    for char in text:
      cat = unicodedata.category(char)
      if cat == "Mn":
        continue
      output.append(char)
    return "".join(output)

  def _run_split_on_punc(self, text):
    """Splits punctuation on a piece of text."""
    chars = list(text)
    i = 0
    start_new_word = True
    output = []
    while i < len(chars):
      char = chars[i]
      if _bert_is_punctuation(char):
        output.append([char])
        start_new_word = True
      else:
        if start_new_word:
          output.append([])
        start_new_word = False
        output[-1].append(char)
      i += 1

    return ["".join(x) for x in output]

  def _tokenize_chinese_chars(self, text):
    """Adds whitespace around any CJK character."""
    output = []
    for char in text:
      cp = ord(char)
      if self._is_chinese_char(cp):
        output.append(" ")
        output.append(char)
        output.append(" ")
      else:
        output.append(char)
    return "".join(output)

  def _is_chinese_char(self, cp):
    """Checks whether CP is the codepoint of a CJK character."""
    # This defines a "chinese character" as anything in the CJK Unicode block:
    #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
    #
    # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
    # despite its name. The modern Korean Hangul alphabet is a different block,
    # as is Japanese Hiragana and Katakana. Those alphabets are used to write
    # space-separated words, so they are not treated specially and handled
    # like the all of the other languages.
    if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
        (cp >= 0x3400 and cp <= 0x4DBF) or  #
        (cp >= 0x20000 and cp <= 0x2A6DF) or  #
        (cp >= 0x2A700 and cp <= 0x2B73F) or  #
        (cp >= 0x2B740 and cp <= 0x2B81F) or  #
        (cp >= 0x2B820 and cp <= 0x2CEAF) or
        (cp >= 0xF900 and cp <= 0xFAFF) or  #
        (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
      return True

    return False

  def _clean_text(self, text):
    """Performs invalid character removal and whitespace cleanup on text."""
    output = []
    for char in text:
      cp = ord(char)
      if cp == 0 or cp == 0xfffd or _bert_is_control(char):
        continue
      if _bert_is_whitespace(char):
        output.append(" ")
      else:
        output.append(char)
    return "".join(output)


class BertWordpieceTokenizer:
  """Runs WordPiece tokenziation."""

  def __init__(self, vocab, unk_token="[UNK]", max_input_chars_per_word=200):
    self.vocab = vocab
    self.unk_token = unk_token
    self.max_input_chars_per_word = max_input_chars_per_word

  def tokenize(self, text):
    """Tokenizes a piece of text into its word pieces.

    This uses a greedy longest-match-first algorithm to perform tokenization
    using the given vocabulary.
    For example:
      input = "unaffable"
      output = ["un", "##aff", "##able"]
    Args:
      text: A single token or whitespace separated tokens. This should have
        already been passed through `BasicTokenizer.

    Returns:
      A list of wordpiece tokens.
    """

    text = native_to_unicode(text)

    output_tokens = []
    for token in whitespace_tokenize(text):
      chars = list(token)
      if len(chars) > self.max_input_chars_per_word:
        output_tokens.append(self.unk_token)
        continue

      is_bad = False
      start = 0
      sub_tokens = []
      while start < len(chars):
        end = len(chars)
        cur_substr = None
        while start < end:
          substr = "".join(chars[start:end])
          if start > 0:
            substr = "##" + substr
          if substr in self.vocab:
            cur_substr = substr
            break
          end -= 1
        if cur_substr is None:
          is_bad = True
          break
        sub_tokens.append(cur_substr)
        start = end

      if is_bad:
        output_tokens.append(self.unk_token)
      else:
        output_tokens.extend(sub_tokens)
    return output_tokens


def _bert_is_whitespace(char):
  """Checks whether `chars` is a whitespace character."""
  # \t, \n, and \r are technically contorl characters but we treat them
  # as whitespace since they are generally considered as such.
  if char == " " or char == "\t" or char == "\n" or char == "\r":
    return True
  cat = unicodedata.category(char)
  if cat == "Zs":
    return True
  return False


def _bert_is_control(char):
  """Checks whether `chars` is a control character."""
  # These are technically control characters but we count them as whitespace
  # characters.
  if char == "\t" or char == "\n" or char == "\r":
    return False
  cat = unicodedata.category(char)
  if cat in ("Cc", "Cf"):
    return True
  return False


def _bert_is_punctuation(char):
  """Checks whether `chars` is a punctuation character."""
  cp = ord(char)
  # We treat all non-letter/number ASCII as punctuation.
  # Characters such as "^", "$", and "`" are not in the Unicode
  # Punctuation class but we treat them as punctuation anyways, for
  # consistency.
  if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
      (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
    return True
  cat = unicodedata.category(char)
  if cat.startswith("P"):
    return True
  return False


def whitespace_tokenize(text):
  """Runs basic whitespace cleaning and splitting on a piece of text."""
  text = text.strip()
  if not text:
    return []
  tokens = text.split()
  return tokens



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.data..tokenizer."""
import os
import random

import six
from six.moves import range  # pylint: disable=redefined-builtin
import tensorflow.compat.v1 as tf
from trax.data import tokenizer


pkg_dir, _ = os.path.split(__file__)
_TESTDATA = os.path.join(pkg_dir, "testdata")


class TokenizerTest(tf.test.TestCase):

  def test_encode(self):
    self.assertListEqual(
        [u"Dude", u" - ", u"that", u"'", u"s", u"so", u"cool", u"."],
        tokenizer.encode(u"Dude - that's so cool."))
    self.assertListEqual([u"Łukasz", u"est", u"né", u"en", u"1981", u"."],
                         tokenizer.encode(u"Łukasz est né en 1981."))
    self.assertListEqual([u" ", u"Spaces", u"at", u"the", u"ends", u" "],
                         tokenizer.encode(u" Spaces at the ends "))
    self.assertListEqual([u"802", u".", u"11b"], tokenizer.encode(u"802.11b"))
    self.assertListEqual([u"two", u". \n", u"lines"],
                         tokenizer.encode(u"two. \nlines"))

  def test_decode(self):
    self.assertEqual(
        u"Dude - that's so cool.",
        tokenizer.decode(
            [u"Dude", u" - ", u"that", u"'", u"s", u"so", u"cool", u"."]))

  def test_invertibility_on_random_strings(self):
    for _ in range(1000):
      s = u"".join(six.unichr(random.randint(0, 65535)) for _ in range(10))
      self.assertEqual(s, tokenizer.decode(tokenizer.encode(s)))


class TestTokenCounts(tf.test.TestCase):

  def setUp(self):
    super(TestTokenCounts, self).setUp()
    self.corpus_path = os.path.join(_TESTDATA, "corpus-*.txt")
    self.vocab_path = os.path.join(_TESTDATA, "vocab-*.txt")

  def test_corpus_token_counts_split_on_newlines(self):
    token_counts = tokenizer.corpus_token_counts(
        self.corpus_path, corpus_max_lines=0, split_on_newlines=True)

    expected = {
        u"'": 2,
        u".": 2,
        u". ": 1,
        u"... ": 1,
        u"Groucho": 1,
        u"Marx": 1,
        u"Mitch": 1,
        u"Hedberg": 1,
        u"I": 3,
        u"in": 2,
        u"my": 2,
        u"pajamas": 2,
    }
    self.assertDictContainsSubset(expected, token_counts)
    self.assertNotIn(u".\n\n", token_counts)
    self.assertNotIn(u"\n", token_counts)

  def test_corpus_token_counts_no_split_on_newlines(self):
    token_counts = tokenizer.corpus_token_counts(
        self.corpus_path, corpus_max_lines=0, split_on_newlines=False)

    self.assertDictContainsSubset({u".\n\n": 2, u"\n": 3}, token_counts)

  def test_corpus_token_counts_split_with_max_lines(self):
    token_counts = tokenizer.corpus_token_counts(
        self.corpus_path, corpus_max_lines=5, split_on_newlines=True)

    self.assertIn(u"slept", token_counts)
    self.assertNotIn(u"Mitch", token_counts)

  def test_corpus_token_counts_no_split_with_max_lines(self):
    token_counts = tokenizer.corpus_token_counts(
        self.corpus_path, corpus_max_lines=5, split_on_newlines=False)

    self.assertIn(u"slept", token_counts)
    self.assertNotIn(u"Mitch", token_counts)
    self.assertDictContainsSubset({
        u".\n\n": 1,
        u"\n": 2,
        u".\n": 1
    }, token_counts)

  def test_vocab_token_counts(self):
    token_counts = tokenizer.vocab_token_counts(self.vocab_path, 0)

    expected = {
        u"lollipop": 8,
        u"reverberated": 12,
        u"kattywampus": 11,
        u"balderdash": 10,
        u"jiggery-pokery": 14,
    }
    self.assertDictEqual(expected, token_counts)

  def test_vocab_token_counts_with_max_lines(self):
    # vocab-1 has 2 lines, vocab-2 has 3
    token_counts = tokenizer.vocab_token_counts(self.vocab_path, 5)

    expected = {
        u"lollipop": 8,
        u"reverberated": 12,
        u"kattywampus": 11,
        u"balderdash": 10,
    }
    self.assertDictEqual(expected, token_counts)


if __name__ == "__main__":
  tf.test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for activation function layers."""

from absl.testing import absltest
import numpy as np

import trax.layers as tl


class ActivationFnsTest(absltest.TestCase):

  def test_relu(self):
    layer = tl.Relu()
    x = np.array([-2.0, -1.0, 0.0, 2.0, 3.0, 5.0])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [0.0, 0.0, 0.0, 2.0, 3.0, 5.0])

  def test_parametric_relu(self):
    layer = tl.ParametricRelu(a=.25)
    x = np.array([-2.0, -1.0, 0.0, 2.0, 3.0, 5.0])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [0.0, 0.0, 0.0, .5, .75, 1.25])

  def test_leaky_relu(self):
    layer = tl.LeakyRelu(a=.125)
    x = np.array([-2.0, -1.0, 0.0, 2.0, 3.0, 5.0])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [-.25, -.125, 0.0, 2.0, 3.0, 5.0])

  def test_hard_sigmoid(self):
    layer = tl.HardSigmoid()
    x = np.array([-1.5, -.5, -.25, 0.0, .25, .5, 1.5])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [0.0, 0.5, 0.75, 1.0, 1.0, 1.0, 1.0])

  def test_hard_tanh(self):
    layer = tl.HardTanh()
    x = np.array([-1.5, -.5, -.25, 0.0, .25, .5, 1.5])
    y = layer(x)
    self.assertEqual(tl.to_list(y), [-1.0, -.5, -.25, 0.0, .25, .5, 1.0])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Layers that compute activation functions.

An activation layer computes element-wise a nonlinear function of the preceding
layer's output. Historically, an activation function was considered part of
each node in each layer of the neural network. Trax follows the common current
practice of separating the activation function as its own layer, which enables
easier experimentation across different activation functions.
"""

from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base
from trax.layers.assert_shape import assert_shape
from trax.layers.base import Fn


@assert_shape('...->...')  # The output and input shapes are the same.
def Relu():
  r"""Returns a layer that computes the Rectified Linear Unit (ReLU) function.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          0 & \text{if}\ x \leq 0, \\
          x & \text{otherwise}.
      \end{array} \right.
  """
  return Fn('Relu', lambda x: jnp.where(x <= 0, jnp.zeros_like(x), x))


@assert_shape('...->...')  # The output and input shapes are the same.
def ParametricRelu(a=1.):
  r"""Returns a layer that computes a ReLU function with the given slope.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          0  & \text{if}\ x \leq 0, \\
          ax & \text{otherwise}.
      \end{array} \right.

  Args:
    a: Slope of line for positive inputs.
  """
  return Fn('ParametricRelu', lambda x: jnp.maximum(a * x, jnp.zeros_like(x)))


@assert_shape('...->...')  # The output and input shapes are the same.
def LeakyRelu(a=0.01):
  r"""Returns a ReLU-like layer with linear nonzero outputs for negative inputs.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          ax & \text{if}\ x \leq 0, \\
          x  & \text{otherwise}.
      \end{array} \right.

  Args:
    a: Slope of line for negative inputs.
  """
  return Fn('LeakyRelu', lambda x: jnp.where(x >= 0, x, a * x))


@assert_shape('...->...')  # The output and input shapes are the same.
def Elu(a=1.):
  r"""Returns a ReLU-like layer with exponential outputs for negative inputs.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          a \cdot (e^x - 1) & \text{if}\ x \leq 0, \\
          x                 & \text{otherwise}.
      \end{array} \right.

  (Asymptotically, :math:`f(x)\rightarrow -a` as :math:`x\rightarrow - \infty`.)

  Args:
    a: Coefficient multiplying the exponential, for negative inputs.
  """
  return Fn('Elu', lambda x: jnp.where(x > 0, x, a * jnp.expm1(x)))


@assert_shape('...->...')  # The output and input shapes are the same.
def Selu(alpha=1.6732632423543772848170429916717,
         lmbda=1.0507009873554804934193349852946):
  r"""Returns an `Elu`-like layer with an additional scaling/slope parameter.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          \lambda \cdot \alpha \cdot (e^x - 1) & \text{if}\ x \leq 0, \\
          \lambda \cdot x                      & \text{otherwise}.
      \end{array} \right.

  Args:
    alpha: Coefficient multiplying the exponential, for negative inputs.
    lmbda: Coefficient scaling the whole function.
  """
  return Fn('Selu', lambda x: lmbda * jnp.where(x > 0, x, alpha * jnp.expm1(x)))


@assert_shape('...->...')  # The output and input shapes are the same.
def Gelu():
  r"""Returns a layer that computes the Gaussian Error Linear Unit function.

  .. math::
      f(x) = \frac{x}{2} \cdot (1 + \hbox{erf}(\frac{x}{\sqrt{2}}))
  """
  return Fn('Gelu', lambda x: x * 0.5 * (1.0 + fastmath.erf(x / jnp.sqrt(2.0))))


@assert_shape('...->...')  # The output and input shapes are the same.
def FastGelu():
  r"""Returns a layer that computes a fast approximation to `Gelu`.

  .. math::
      f(x) = \frac{x}{2} \cdot (1 + \tanh(ax + abx^3))

  where :math:`a = 0.7978845608` and :math:`b = 0.044715`.
  """
  def f(x):  # pylint: disable=invalid-name
    return 0.5 * x * (1 + jnp.tanh(x * 0.7978845608 * (1 + 0.044715 * x * x)))
  return Fn('FastGelu', f)


# pylint: disable=unnecessary-lambda
@assert_shape('...->...')  # The output and input shapes are the same.
def Sigmoid():
  r"""Returns a layer that computes the sigmoid function.

  .. math::
      f(x) = \frac{1}{1 + e^{-x}}
  """
  return Fn('Sigmoid', lambda x: fastmath.expit(x))


@assert_shape('...->...')  # The output and input shapes are the same.
def Tanh():
  r"""Returns a layer that computes the hyperbolic tangent function.

  .. math::
      f(x) = \frac{e^x - e^{-x}}{e^x + e^{-x}}
  """
  return Fn('Tanh', lambda x: jnp.tanh(x))
# pylint: enable=unnecessary-lambda


@assert_shape('...->...')  # The output and input shapes are the same.
def HardSigmoid():
  r"""Returns a layer that computes a linear approximation to `Sigmoid`.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          0 & \text{if}\ x \leq 0, \\
          x & \text{if}\ 0 < x < 1, \\
          1 & \text{otherwise}.
      \end{array} \right.
  """
  return Fn('HardSigmoid', lambda x: jnp.maximum(0, jnp.minimum(1, (1 + x))))


@assert_shape('...->...')  # The output and input shapes are the same.
def HardTanh():
  r"""Returns a layer that computes a linear approximation to `Tanh`.

  .. math::
      f(x) = \left\{ \begin{array}{cl}
          -1 & \text{if}\ x \leq -1, \\
          x  & \text{if}\ -1 < x < 1, \\
          1  & \text{otherwise}.
      \end{array} \right.
  """
  return Fn('HardTanh', lambda x: jnp.maximum(-1, jnp.minimum(1, x)))


@assert_shape('...->...')  # The output and input shapes are the same.
def Softplus():
  r"""Returns a layer that computes the softplus function.

  .. math::
      f(x) = \ln(e^x + 1)
  """
  return Fn('Softplus', lambda x: jnp.logaddexp(x, 0.))


@assert_shape('...->...')  # The output and input shapes are the same.
def Exp():
  """Returns a layer that computes the element-wise exponential of a tensor."""
  return Fn('Exp', lambda x: jnp.exp(x))  # pylint: disable=unnecessary-lambda


@assert_shape('...->...')  # The output and input shapes are the same.
def Log():
  """Returns a layer that computes the element-wise logarithm of a tensor."""
  return Fn('Log', lambda x: jnp.log(x))  # pylint: disable=unnecessary-lambda


@assert_shape('...->...')  # The output and input shapes are the same.
def Swish():
  r"""Returns a layer that computes the Swish function.

  .. math::
      f(x) = x \cdot \text{sigmoid}(x)
  """
  return Fn('Swish', lambda x: x * fastmath.expit(x))


@assert_shape('...a->...b')  # The output and input shapes are not the same.
def Glu():
  r"""Returns a layer that computes the Gated Linear Unit function.

  .. math::
      f(x) = a \cdot \text{sigmoid}(b)
  where a and b are formed by splitting input in half along axis

  """

  def _f(x, axis=-1):  # pylint: disable=invalid-name
    size = x.shape[axis]
    assert size % 2 == 0, f'axis {axis} of size {size} is not be divisible by 2'
    a, b = jnp.split(x, 2, axis)
    return a * fastmath.expit(b)

  return Fn('Glu', _f)


class ThresholdedLinearUnit(base.Layer):
  """Thresholded Linear Unit, c.f. https://arxiv.org/pdf/1911.09737.pdf ."""

  def init_weights_and_state(self, input_signature):
    """Initializes this layer's single weight to zero."""
    del input_signature
    self.weights = jnp.zeros((), dtype=jnp.float32)

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model.

    Args:
      inputs: Tensor.

    Returns:
      Tensor of same shape and dtype as the input.
    """
    threshold = self.weights
    return jnp.maximum(inputs, threshold)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for core layers."""

from absl.testing import absltest
import numpy as np

from trax import shapes
from trax.fastmath import numpy as jnp
import trax.layers as tl
import trax.layers.initializers as init


class DenseTest(absltest.TestCase):
  """Test Dense layer per se and as a key example of trainable layers."""

  def test_call_before_init_raises_error(self):
    layer = tl.Dense(5)
    x = np.array([1, 2, 3])

    # Without init, layer lacks the weights it needs for forward computation.
    with self.assertRaises(tl.LayerError):
      _ = layer(x)

  def test_call_uses_and_caches_supplied_weights(self):
    layer = tl.Dense(4)
    x = np.array([2, 3])

    # Weights from random initialization are cached in the layer.
    _, _ = layer.init(shapes.signature(x))
    w_init, b_init = layer.weights

    # Call the layer with externally specified weights.
    w = np.array([[10000, 20000, 30000, 40000], [100, 200, 100, 200]])
    b = np.array([9, 8, 7, 6])
    y = layer(x, weights=(w, b))

    # Using weights keyword arg overrides any previous cached weights ...
    self.assertEqual(y.tolist(), [20309, 40608, 60307, 80606])
    self.assertNotEqual(w.tolist(), w_init.tolist())
    self.assertNotEqual(b.tolist(), b_init.tolist())

    # ... and do not over-write the old weights.
    w_cached, b_cached = layer.weights
    self.assertNotEqual(w.tolist(), w_cached.tolist())
    self.assertNotEqual(b.tolist(), b_cached.tolist())

  def test_separate_instances_have_separate_weights(self):
    # Two dense layer instances: each will get its own initial weights (w, b).
    model = tl.Serial(tl.Dense(5), tl.Dense(5))

    sample_input = np.array([1, 2, 3, 4, 5])
    _, _ = model.init(shapes.signature(sample_input))
    weights_0 = model.sublayers[0].weights
    weights_1 = model.sublayers[1].weights

    w0, b0 = weights_0
    w1, b1 = weights_1
    self.assertNotEqual(w0.tolist(), w1.tolist())
    self.assertNotEqual(b0.tolist(), b1.tolist())

  def test_shared_instance_means_shared_weights(self):
    # Same dense layer instance in two places --> shared weights.
    layer = tl.Dense(5)
    model = tl.Serial(layer, layer)
    sample_input = np.array([1, 2, 3, 4, 5])
    weights, _ = model.init(shapes.signature(sample_input))
    self.assertIs(weights[1], tl.GET_WEIGHTS_FROM_CACHE)

  def test_call_no_bias(self):
    layer = tl.Dense(4, use_bias=False)
    x = np.array([2, 5, 3])
    _, _ = layer.init(shapes.signature(x))

    w = np.array([[100, 200, 300, 400], [10, 10, 10, 10], [1, 2, 1, 2]])
    y = layer(x, weights=w)
    self.assertEqual(y.tolist(), [253, 456, 653, 856])

  def test_new_weights_use_bias(self):
    layer = tl.Dense(4)
    x = np.array([1, 2])
    _, _ = layer.init(shapes.signature(x))
    self.assertLen(layer.weights, 2)
    self.assertEqual(layer.weights[0].shape, (2, 4))
    self.assertEqual(layer.weights[1].shape, (4,))

  def test_new_weights_no_bias(self):
    layer = tl.Dense(4, use_bias=False)
    x = np.array([1, 2])
    _, _ = layer.init(shapes.signature(x))
    self.assertEqual(layer.weights.shape, (2, 4))

  def test_init_twice_weights_same_shape(self):
    layer = tl.Dense(4, use_bias=False)
    x = np.array([1, 2])
    w1, _ = layer.init(shapes.signature(x))
    w2, _ = layer.init(shapes.signature(x))
    self.assertEqual(w1.shape, (2, 4))
    self.assertEqual(w2.shape, (2, 4))

  def test_save_to_file_and_init_to_file(self):
    layer1 = tl.Dense(4, use_bias=False)
    layer2 = tl.Dense(4, use_bias=False)
    x = np.array([1, 2])
    w1, _ = layer1.init(shapes.signature(x))
    layer1.save_to_file('/tmp/dense_weights',
                        input_signature=shapes.signature(x))
    w2, _ = layer2.init_from_file('/tmp/dense_weights')
    self.assertEqual(w1.shape, (2, 4))
    self.assertEqual(w2.shape, (2, 4))
    self.assertEqual(w1.tolist(), w2.tolist())


class EmbeddingTest(absltest.TestCase):

  def test_forward(self):
    layer = tl.Embedding(10, 3)  # vocab_size=10, d_feature=3
    _, _ = layer.init(None)  # Embedding init doesn't use input signature.
    x = np.array([2, 3, 5, 3, 2])
    y = layer(x)
    self.assertEqual(y.shape, (5, 3))

    # For distinct in-domain token IDs, resulting vectors should be distinct.
    self.assertNotEqual(y[0].tolist(), y[1].tolist())
    self.assertNotEqual(y[0].tolist(), y[2].tolist())
    self.assertNotEqual(y[1].tolist(), y[2].tolist())

    # For repeats of a token id, resulting vectors should match.
    self.assertEqual(y[0].tolist(), y[4].tolist())
    self.assertEqual(y[1].tolist(), y[3].tolist())

  def test_negative_inputs_clip_to_zero(self):
    layer = tl.Embedding(10, 3)
    _, _ = layer.init(None)
    x = np.array([0, 2, 3, -2, -3])
    y = layer(x)
    self.assertNotEqual(y[0].tolist(), y[1].tolist())
    self.assertNotEqual(y[0].tolist(), y[2].tolist())
    self.assertEqual(y[0].tolist(), y[3].tolist())
    self.assertEqual(y[0].tolist(), y[4].tolist())

  def test_large_inputs_clip_to_upper_bound(self):
    layer = tl.Embedding(10, 3)
    _, _ = layer.init(None)
    x = np.array([2, 3, 9, 10, 20])
    y = layer(x)

    # vocab_size of 10 means max valid token id is 9.
    self.assertNotEqual(y[2].tolist(), y[0].tolist())
    self.assertNotEqual(y[2].tolist(), y[1].tolist())
    self.assertEqual(y[2].tolist(), y[3].tolist())
    self.assertEqual(y[2].tolist(), y[4].tolist())

  def test_new_weights(self):
    layer = tl.Embedding(20, 5)
    _, _ = layer.init(None)

    # Default weights sampled from Gaussian, mu = 0, sigma = 1.
    w = layer.weights
    self.assertEqual(w.shape, (20, 5))
    self.assertLess(np.abs(np.mean(w)), .4)  # .4 is 4 sigma deviation

  def test_explicit_kernel_initializer(self):

    def f(shape, rng):
      del rng
      n_elements = np.prod(shape)
      return np.arange(n_elements).reshape(shape)

    layer = tl.Embedding(5, 2, kernel_initializer=f)
    _, _ = layer.init(None)
    x = np.array([0, 1, 2, 3, 4])
    y = layer(x)
    self.assertEqual(y.tolist(), [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])


class DropoutTest(absltest.TestCase):

  def test_call_in_train_mode(self):
    layer = tl.Dropout(rate=0.1, mode='train')
    x = np.ones((2, 5, 1000))  # 10,000 values
    y = layer(x)
    self.assertEqual(y.shape, (2, 5, 1000))

    # Dropout is stochastic; test it nonflakily at 4 sigmas (.99994).
    n_remaining = np.count_nonzero(y)
    mu_of_remaining = 9000  # N * q:  10000 * .9
    sigma_of_remaining = 30  # sqrt(N * p * q):  sqrt(10000 * .1 * .9)
    self.assertLess(
        np.abs(n_remaining - mu_of_remaining), 4 * sigma_of_remaining)

  def test_call_in_eval_mode_does_no_dropout(self):
    layer = tl.Dropout(rate=0.1, mode='eval')
    x = np.ones((2, 5, 1000))
    y = layer(x)
    self.assertEqual(np.count_nonzero(y), 10_000)

  def test_new_weights(self):
    layer = tl.Dropout(rate=0.1, mode='train')
    layer.init(None)
    self.assertEmpty(layer.weights)


class WeightsTest(absltest.TestCase):
  """Test Weights layer."""

  def test_simple(self):
    layer = tl.Weights(lambda shape, rng: jnp.zeros(shape, dtype=jnp.float32))
    layer.init(())
    y = layer(())
    self.assertEqual(y.tolist(), 0.)

  def test_shape(self):
    layer = tl.Weights(init.RandomNormalInitializer(), (5, 10, 3))
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, (5, 10, 3))

  def test_simple_custom_initializer(self):
    layer = tl.Weights(init.RandomNormalInitializer())
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, ())
    self.assertNotEqual(y.tolist(), 0.)

  def test_custom_initializer_shape(self):
    layer = tl.Weights(lambda shape, rng: jnp.zeros(shape, dtype=jnp.float32),
                       (2, 2))
    layer.init(())
    y = layer(())
    self.assertEqual(y.tolist(), [[0., 0.],
                                  [0., 0.]])

    layer = tl.Weights(init.RandomNormalInitializer(), (2, 2))
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, (2, 2))
    self.assertNotEqual(y.tolist(), [[0., 0.],
                                     [0., 0.]])


class SummaryScalarTest(absltest.TestCase):

  def test_passes(self):
    layer = tl.SummaryScalar('test')
    x = np.array([[3., 5.], [2., 6.]])  # 10,000 values
    y = layer(x)
    self.assertEqual(y.tolist(), [[3., 5.], [2., 6.]])
    self.assertEqual(layer.state['summary_test'].tolist(), 4.0)


class RandomUniformTest(absltest.TestCase):
  """Test Weights layer."""

  def test_simple(self):
    layer = tl.RandomUniform()
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, ())
    self.assertBetween(y, 0.0, 1.0)

  def test_shape(self):
    layer = tl.RandomUniform(shape=(5, 10, 3))
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, (5, 10, 3))

  def test_simple_range(self):
    layer = tl.RandomUniform(1., 2., shape=(1000,))
    layer.init(())
    y = layer(())
    self.assertEqual(y.shape, (1000,))
    self.assertBetween(min(y.tolist()), 1., 2.)
    self.assertBetween(max(y.tolist()), 1., 2.)
    self.assertBetween(1.5, min(y.tolist()), max(y.tolist()))


class LocallyConnected1dTest(absltest.TestCase):

  def test_shape_kernel1(self):
    for padding in ['WRAP', 'SAME', 'VALID']:
      layer = tl.LocallyConnected1d(6, 1, padding=padding)
      x = np.array([[0, 1], [2, 3], [4, 5]])
      layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, (3, 6))

  def test_shape_kernel3(self):
    for padding in ['WRAP', 'SAME']:
      layer = tl.LocallyConnected1d(6, 3, padding=padding)
      x = np.array([[0, 1], [2, 3], [4, 5]])
      layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, (3, 6))

    for padding in ['VALID']:
      layer = tl.LocallyConnected1d(6, 3, padding=padding)
      x = np.array([[0, 1], [2, 3], [4, 5]])
      layer.init(shapes.signature(x))
      y = layer(x)
      self.assertEqual(y.shape, (1, 6))


class FlattenTest(absltest.TestCase):

  def test_keep_default(self):
    layer = tl.Flatten()
    x = np.ones((1, 2, 3, 4, 5))
    y = layer(x)
    # Default is leave first axis untouched, flatten the rest.
    self.assertEqual(y.shape, (1, 2 * 3 * 4 * 5))

  def test_keep_3(self):
    layer = tl.Flatten(n_axes_to_keep=3)
    x = np.ones((1, 2, 3, 4, 5))
    y = layer(x)
    self.assertEqual(y.shape, (1, 2, 3, 4 * 5))

  def test_keep_max_number(self):
    layer = tl.Flatten(n_axes_to_keep=4)
    x = np.ones((1, 2, 3, 4, 5))
    y = layer(x)
    self.assertEqual(y.shape, (1, 2, 3, 4, 5))

  def test_keep_too_many_raises_error(self):
    layer = tl.Flatten(n_axes_to_keep=5)
    with self.assertRaises(tl.LayerError):
      x = np.ones((1, 2, 3, 4, 5))
      _ = layer(x)


class LogSoftmaxTest(absltest.TestCase):

  def test_call(self):
    layer = tl.LogSoftmax()
    x = np.array([[2., 1., -10.],
                  [1., 1., -10.]])
    y = layer(x)
    np.testing.assert_allclose(y,
                               [[-0.313, -1.313, -12.313],
                                [-0.693, -0.693, -11.693]],
                               atol=.001)


class SoftmaxTest(absltest.TestCase):

  def test_call(self):
    layer = tl.Softmax()
    x = np.array([[2., 1., -10.],
                  [1., 1., -10.]])
    y = layer(x)
    np.testing.assert_allclose(y,
                               [[.731, .269, .00000449],
                                [.500, .500, .00000835]],
                               atol=.001)


class CoreFunctionsTest(absltest.TestCase):

  def test_one_hot(self):
    targets = np.array([2, 0, 1])
    n_categories = 5
    target_distributions = tl.one_hot(targets, n_categories)
    self.assertEqual(tl.to_list(target_distributions),
                     [[0., 0., 1., 0., 0.],
                      [1., 0., 0., 0., 0.],
                      [0., 1., 0., 0., 0.]])

  def test_log_softmax(self):
    activations = np.array([[2., 1., -10.],
                            [1., 1., -10.]])
    log_probabilities = tl.log_softmax(activations)
    np.testing.assert_allclose(log_probabilities,
                               [[-0.313, -1.313, -12.313],
                                [-0.693, -0.693, -11.693]],
                               atol=.001)

  def test_log_gaussian_pdf(self):
    x = np.zeros((2, 5), dtype=np.float32)
    mu = x
    dsigma = np.eye(5)[None, :, :]
    sigma = np.concatenate([dsigma, 2 * dsigma], axis=0)
    prob = tl.log_gaussian_pdf(x, mu, sigma)
    self.assertEqual(prob.shape, (2,))
    self.assertEqual(int(prob[0]), -4)
    self.assertEqual(int(prob[1]), -6)

  def test_log_gaussian_diag_pdf(self):
    x = np.zeros((2, 5), dtype=np.float32)
    mu = x
    sigma = np.ones((5,))[None, :]
    sigma = np.concatenate([sigma, 2 * sigma], axis=0)
    prob = tl.log_gaussian_diag_pdf(x, mu, sigma)
    self.assertEqual(prob.shape, (2,))
    self.assertEqual(int(prob[0]), -4)
    self.assertEqual(int(prob[1]), -6)


class StopGradientTest(absltest.TestCase):

  def test_passes(self):
    layer = tl.StopGradient()
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2, 2))
    self.assertEqual(y.tolist(), [[3., 5.], [2., 6.]])


class MinMaxTest(absltest.TestCase):

  def test_min(self):
    layer = tl.Min()
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2,))
    self.assertEqual(y.tolist(), [3., 2.])

    layer = tl.Min(axis=0)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2,))
    self.assertEqual(y.tolist(), [2., 5.])

    layer = tl.Min(axis=None)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, ())
    self.assertEqual(y.tolist(), 2.)

    layer = tl.Min(keepdims=True)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2, 1))
    self.assertEqual(y.tolist(), [[3.], [2.]])

  def test_max(self):
    layer = tl.Max()
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2,))
    self.assertEqual(y.tolist(), [5., 6.])

    layer = tl.Max(axis=0)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (2,))
    self.assertEqual(y.tolist(), [3., 6.])

    layer = tl.Max(axis=None)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, ())
    self.assertEqual(y.tolist(), 6.)

    layer = tl.Max(axis=0, keepdims=True)
    x = np.array([[3., 5.], [2., 6.]])
    y = layer(x)
    self.assertEqual(y.shape, (1, 2))
    self.assertEqual(y.tolist(), [[3., 6.]])


class ClassifierLayersTest(absltest.TestCase):

  def test_threshold_to_binary(self):
    layer = tl.ThresholdToBinary()
    x = np.array([.30, .49, .50, .51, .70])
    y = layer(x)
    self.assertEqual(y.tolist(), [0, 0, 0, 1, 1])

  def test_arg_max(self):
    layer = tl.ArgMax()
    x = np.array([[.10, .90, .20, .80],
                  [.22, .88, .11, .99]])
    y = layer(x)
    self.assertEqual(y.tolist(), [1, 3])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax fast math: TF backend."""

import numpy as np
import tensorflow.compat.v2 as tf

from trax.shapes import ShapeDtype
from trax.tf_numpy import extensions as tf_np_extensions
from trax.tf_numpy import numpy as tf_np


def tf_abstract_eval(f):
  """Returns a function that evaluates `f` given input shapes and dtypes.

  It transforms function `f` to a function that performs the same computation as
  `f` but only on shapes and dtypes (a.k.a. shape inference).

  Args:
    f: the function to be transformed.

  Returns:
    A function whose input arguments can be either the same as `f`'s or only
    their shapes/dtypes represented by `ShapeDtype`, and whose return values are
    `ShapeDtype`s with the same nested structure as `f`'s return values.
  """
  f_shape = tf_np_extensions.eval_on_shapes(f)
  def from_shape_type(x):
    if isinstance(x, ShapeDtype):
      return tf.TensorSpec(x.shape, x.dtype)
    else:
      return x
  def to_shape_type(x):  # pylint: disable=missing-docstring
    # TODO(wangpeng): handle partial output shapes using `tf.shape`.
    def to_numpy_shape(s):
      if s.is_fully_defined():
        return tuple(s.as_list())
      else:
        raise ValueError("The output shapes (%s) of the dry-run'ed function are"
                         ' not fully defined.' % s)
    def to_numpy_dtype(t):
      return np.dtype(t.as_numpy_dtype)
    if isinstance(x, tf.TensorSpec):
      return ShapeDtype(to_numpy_shape(x.shape), to_numpy_dtype(x.dtype))
    else:
      return x
  def f_return(*args):
    args = tf.nest.map_structure(from_shape_type, args)
    res = f_shape(*args)
    return tf.nest.map_structure(to_shape_type, res)
  return f_return


# The arguments order is different from tf_np_extensions.uniform
def tf_randint(key, shape, minval, maxval, dtype=np.int32):
  """Sample uniform random values in [minval, maxval) with given shape/dtype.

  Args:
    key: a PRNGKey used as the random key.
    shape: a tuple of nonnegative integers representing the shape.
    minval: int or array of ints broadcast-compatible with ``shape``, a minimum
      (inclusive) value for the range.
    maxval: int or array of ints broadcast-compatible with  ``shape``, a maximum
      (exclusive) value for the range.
    dtype: optional, an int dtype for the returned values (default int32).

  Returns:
    A random array with the specified shape and dtype.
  """
  return tf_np_extensions.uniform(key, shape, minval=minval, maxval=maxval,
                                  dtype=dtype)


_tf_xla_forced_compile_enabled = False


def tf_xla_forced_compile_enabled():
  return _tf_xla_forced_compile_enabled


def set_tf_xla_forced_compile(b):
  global _tf_xla_forced_compile_enabled
  _tf_xla_forced_compile_enabled = b


def _tf_jit(*args, **kwargs):
  kwargs['xla_forced_compile'] = tf_xla_forced_compile_enabled()
  kwargs.pop('donate_argnums', None)  # donate_argnums not used in TF
  return tf_np_extensions.jit(*args, **kwargs)


def _tf_pmap(*args, **kwargs):
  kwargs.pop('donate_argnums', None)  # donate_argnums not used in TF
  return tf_np_extensions.pmap(*args, **kwargs)


def _tf_grad(f, **kwargs):
  """Grad with support for argnums."""
  argnums = kwargs.pop('argnums', 0)
  if argnums != 0:
    def g(*args, **kwargs):
      args = list(args)
      args[0], args[argnums] = args[argnums], args[0]
      return f(*args, **kwargs)
  else:
    g = f
  grad_g = tf_np_extensions.grad(g, **kwargs)
  if argnums == 0:
    return grad_g
  def grad_f(*args, **kwargs):
    args = list(args)
    args[0], args[argnums] = args[argnums], args[0]
    return grad_g(*args, **kwargs)
  return grad_f


def _fold_in(rng, d):
  """Equivalent of jax.random.fold_in."""
  # TODO(lukaszkaiser): verify that this function has good randomness
  # properties or switch to an implementation equivalent to JAX.
  _, rng = tf_np_extensions.split(rng + tf_np.sum(d).astype(tf_np.int64), 2)
  return rng


TF_BACKEND = {
    'name': 'tensorflow-numpy',
    'np': tf_np,
    'jit': _tf_jit,
    'stop_gradient': tf_np_extensions.stop_gradient,
    'grad': _tf_grad,
    'vjp': tf_np_extensions.vjp,
    'custom_grad': tf_np_extensions.custom_grad,
    'abstract_eval': tf_abstract_eval,
    'expit': tf_np_extensions.expit,
    'erf': tf_np_extensions.erf,
    'index_update': tf_np_extensions.index_update,
    'index_add': tf_np_extensions.index_add,
    'index_min': tf_np_extensions.index_min,
    'index_max': tf_np_extensions.index_max,
    'dynamic_slice': tf_np_extensions.dynamic_slice,
    'dynamic_slice_in_dim': tf_np_extensions.dynamic_slice_in_dim,
    'dynamic_update_slice': tf_np_extensions.dynamic_update_slice,
    'dynamic_update_slice_in_dim': tf_np_extensions.dynamic_update_slice_in_dim,
    'logsumexp': tf_np_extensions.logsumexp,
    'conv': tf_np_extensions.conv,
    'lt': lambda x, y: x < y,
    'avg_pool': tf_np_extensions.avg_pool,
    'max_pool': tf_np_extensions.max_pool,
    'sort_key_val': tf_np_extensions.sort_key_val,
    'random_uniform': tf_np_extensions.uniform,
    'random_randint': tf_randint,
    'random_normal': tf_np_extensions.normal,
    'random_bernoulli': tf_np_extensions.bernoulli,
    'random_get_prng': tf_np_extensions.prng,
    'random_split': tf_np_extensions.split,
    'random_fold_in': _fold_in,
    # TODO(wangpeng): See whether and how to support `remat`
    'remat': lambda f: f,
    'scan': tf_np_extensions.scan,
    'map': tf_np_extensions.tf_map,
    # TODO(wangpeng): can we make extensions ds_as_numpy compatible with data?
    # 'dataset_as_numpy': tf_np_extensions.dataset_as_numpy,
    'global_device_count': lambda: max(len(tf_np_extensions.accelerators()), 1),
    'local_device_count': lambda: max(len(tf_np_extensions.accelerators()), 1),
    'pmap': _tf_pmap,
    'psum': tf_np_extensions.psum,
    'vmap': tf_np_extensions.vmap,
}



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.rl.advantages."""

import functools

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from trax.rl import advantages


def calc_bias_and_variance(x, true_mean):
  sample_mean = np.mean(x)
  bias = np.mean(np.abs(sample_mean - true_mean))
  variance = np.mean((x - sample_mean) ** 2)
  return (bias, variance)


def estimate_advantage_bias_and_variance(
    advantage_fn,
    mean_reward=1.23,
    reward_noise=0.45,
    discount_mask=None,
    discount_true_return=True,
    true_value=False,
    n_samples=10000,
    length=5,
    gamma=0.9,
    margin=1,
    **advantage_kwargs
):
  advantage_fn = advantage_fn(gamma, margin, **advantage_kwargs)
  rewards = np.random.normal(
      loc=mean_reward, scale=reward_noise, size=(n_samples, length)
  )
  if discount_mask is None:
    discount_mask = np.ones_like(rewards)
  gammas = advantages.mask_discount(gamma, discount_mask)
  returns = advantages.discounted_returns(rewards, gammas)

  true_returns = advantages.discounted_returns(
      np.full(returns.shape, fill_value=mean_reward), gammas=gammas
  )
  if true_value:
    values = true_returns
  else:
    values = np.zeros_like(returns)

  dones = np.zeros_like(returns, dtype=np.bool)
  adv = advantage_fn(rewards, returns, values, dones, discount_mask)
  if discount_true_return:
    mean_return = true_returns[0, 0]
  else:
    mean_return = mean_reward * length
  return calc_bias_and_variance(adv[:, 0], mean_return - values[:, 0])


class AdvantagesTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('monte_carlo', advantages.monte_carlo),
      ('td_k', advantages.td_k),
      ('td_lambda', advantages.td_lambda),
      ('gae', advantages.gae),
  )
  def test_shapes(self, advantage_fn):
    rewards = np.array([[1, 1, 1]], dtype=np.float32)
    returns = np.array([[3, 2, 1]], dtype=np.float32)
    values = np.array([[2, 2, 2]], dtype=np.float32)
    dones = np.array([[False, False, True]])
    discount_mask = np.ones_like(rewards)
    adv1 = advantage_fn(gamma=1, margin=1)(
        rewards, returns, values, dones, discount_mask
    )
    self.assertEqual(adv1.shape, (1, 2))
    adv2 = advantage_fn(gamma=1, margin=2)(
        rewards, returns, values, dones, discount_mask
    )
    self.assertEqual(adv2.shape, (1, 1))

  def test_monte_carlo_bias_is_zero(self):
    (bias, _) = estimate_advantage_bias_and_variance(
        advantages.monte_carlo, margin=3
    )
    np.testing.assert_allclose(bias, 0, atol=0.1)

  def test_td_k_variance_lower_than_monte_carlo(self):
    (_, var_td_3) = estimate_advantage_bias_and_variance(
        advantages.td_k, margin=3
    )
    (_, var_mc) = estimate_advantage_bias_and_variance(advantages.monte_carlo)
    self.assertLess(var_td_3, var_mc)

  @parameterized.named_parameters(('1_2', 1, 2), ('2_3', 2, 3))
  def test_td_k_bias_decreases_with_k(self, k1, k2):
    (bias1, _) = estimate_advantage_bias_and_variance(
        advantages.td_k, margin=k1
    )
    (bias2, _) = estimate_advantage_bias_and_variance(
        advantages.td_k, margin=k2
    )
    self.assertGreater(bias1, bias2)

  @parameterized.named_parameters(('1_2', 1, 2), ('2_3', 2, 3))
  def test_td_k_variance_increases_with_k(self, k1, k2):
    (_, var1) = estimate_advantage_bias_and_variance(
        advantages.td_k, margin=k1
    )
    (_, var2) = estimate_advantage_bias_and_variance(
        advantages.td_k, margin=k2
    )
    self.assertLess(var1, var2)

  def test_td_lambda_variance_lower_than_monte_carlo(self):
    (_, var_td_095) = estimate_advantage_bias_and_variance(
        advantages.td_lambda, lambda_=0.95
    )
    (_, var_mc) = estimate_advantage_bias_and_variance(advantages.monte_carlo)
    self.assertLess(var_td_095, var_mc)

  @parameterized.named_parameters(
      ('td_lambda_0.5_0.7', advantages.td_lambda, 0.5, 0.7),
      ('td_lambda_0.7_0.9', advantages.td_lambda, 0.7, 0.9),
      ('gae_0.5_0.7', advantages.gae, 0.5, 0.7),
      ('gae_0.7_0.9', advantages.gae, 0.7, 0.9),
  )
  def test_bias_decreases_with_lambda(self, advantage_fn, lambda1, lambda2):
    (bias1, _) = estimate_advantage_bias_and_variance(
        advantage_fn, lambda_=lambda1
    )
    (bias2, _) = estimate_advantage_bias_and_variance(
        advantage_fn, lambda_=lambda2
    )
    self.assertGreater(bias1, bias2)

  @parameterized.named_parameters(('0.5_0.7', 0.5, 0.7), ('0.7_0.9', 0.7, 0.9))
  def test_variance_increases_with_lambda(self, lambda1, lambda2):
    (_, var1) = estimate_advantage_bias_and_variance(
        advantages.td_lambda, lambda_=lambda1
    )
    (_, var2) = estimate_advantage_bias_and_variance(
        advantages.td_lambda, lambda_=lambda2
    )
    self.assertLess(var1, var2)

  @parameterized.named_parameters(
      ('monte_carlo', advantages.monte_carlo),
      ('td_k', advantages.td_k),
      ('td_lambda', advantages.td_lambda),
      ('gae', advantages.gae),
  )
  def test_advantage_future_return_is_zero_at_done(self, advantage_fn):
    rewards = np.array([[1, 1, 1]], dtype=np.float32)
    returns = np.array([[3, 2, 1]], dtype=np.float32)
    values = np.array([[2, 2, 2]], dtype=np.float32)
    dones = np.array([[False, True, False]])
    discount_mask = np.ones_like(rewards)
    adv = advantage_fn(gamma=0.9, margin=1)(
        rewards, returns, values, dones, discount_mask
    )
    target_returns = values[:, :-1] + adv
    # Assert that in the "done" state the future return in the advantage is
    # zero, i.e. the return is equal to the reward.
    np.testing.assert_almost_equal(target_returns[0, 1], rewards[0, 1])

  @parameterized.named_parameters(
      ('monte_carlo', advantages.monte_carlo),
      # Disabled for TD-k because the differences are too small.
      # ('td_k', advantages.td_k),
      ('td_lambda', advantages.td_lambda),
      ('gae', advantages.gae),
  )
  def test_bias_and_variance_with_non_const_discount_mask(self, advantage_fn):
    non_const_discount_mask = np.array([[1, 0, 1, 0, 1]])
    const_discount_mask = np.ones_like(non_const_discount_mask)
    est_bias_and_variance = functools.partial(
        estimate_advantage_bias_and_variance,
        advantage_fn,
        length=const_discount_mask.shape[1],
        # Set gamma to a small value to accentuate the differences.
        gamma=0.5,
        # We want to measure error due to the discount, so compare with the
        # undiscounted return.
        discount_true_return=False,
        # Use true values to remove the value estimation error.
        true_value=True,
    )
    (bias_non_const, var_non_const) = est_bias_and_variance(
        discount_mask=non_const_discount_mask
    )
    (bias_const, var_const) = est_bias_and_variance(
        discount_mask=const_discount_mask
    )
    self.assertLess(bias_non_const, bias_const)
    self.assertGreater(var_non_const, var_const)

  @parameterized.named_parameters(
      ('monte_carlo', advantages.monte_carlo),
      ('td_k', advantages.td_k),
      ('td_lambda', advantages.td_lambda),
      ('gae', advantages.gae),
  )
  def test_future_return_is_zero_iff_discount_mask_is_on(self, advantage_fn):
    # (... when gamma=0)
    rewards = np.array([[1, 2, 3, 4]], dtype=np.float32)
    values = np.array([[5, 6, 7, 8]], dtype=np.float32)
    dones = np.zeros_like(rewards, dtype=np.bool)
    discount_mask = np.array([[1, 0, 1, 0]], dtype=np.bool)
    gammas = advantages.mask_discount(0.0, discount_mask)
    returns = advantages.discounted_returns(rewards, gammas)
    adv = advantage_fn(gamma=0.0, margin=1)(
        rewards, returns, values, dones, discount_mask
    )
    target_returns = values[:, :-1] + adv
    # Assert that in the states with discount_mask on the future return in the
    # advantage is zero, i.e. the return is equal to the reward.
    rewards = rewards[:, :-1]
    discount_mask = discount_mask[:, :-1]
    np.testing.assert_almost_equal(
        target_returns[discount_mask], rewards[discount_mask]
    )
    # Assert the converse.
    with np.testing.assert_raises(AssertionError):
      np.testing.assert_almost_equal(
          target_returns[~discount_mask], rewards[~discount_mask]
      )


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""The key layer abstraction (Layer class) and supporting machinery."""

import copy
import functools
import gzip
import inspect
import pickle
import random
import traceback

import jax
import numpy as np
import tensorflow as tf

from trax import fastmath
from trax.fastmath import nested_map
from trax.fastmath import numpy as jnp
from trax.shapes import ShapeDtype
from trax.shapes import signature


# TODO(lukaszkaiser): should we use special objects for these for clarity?
EMPTY_WEIGHTS = ()    # Used for layers that have no trainable weights.
EMPTY_STATE = ()      # Used for layers that have no non-trainable state.
GET_WEIGHTS_FROM_CACHE = {'__marker_for_cached_weights_': ()}
GET_STATE_FROM_CACHE = {'__marker_for_cached_state_': ()}
N_WEIGHTS_SHARDS = 1  # TODO(lukaszkaiser): make weight-sharding non-global


class Layer:
  """Base class for composable layers in a deep learning network.

  Layers are the basic building blocks for deep learning models. A layer
  computes a function from zero or more inputs to zero or more outputs,
  optionally using trainable weights (common) and non-parameter state (not
  common).

  Layer subclasses typically override at most two methods of the base `Layer`
  class:

    `forward(inputs)`:
      Computes the layer's output as part of a forward pass through the model.

    `init_weights_and_state(self, input_signature)`:
      Initializes the layer's weights and state to handle input with the given
      signature (number, shapes and dtypes of input arguments).

  A small number of layer types are combinators -- they organize the computation
  of their sublayers, e.g., applying their sublayers in series or in parallel.

  All layers have the following properties, with default values implemented
  in the base `Layer` class:

    - `n_in`: int (default 1)
    - `n_out`: int (default 1)
    - `weights`: tuple (default empty -- the layer has no weights)
    - `state`: tuple (default empty -- the layer has no non-parameter state)
    - `sublayers`: tuple (default empty -- the layer has no sublayers)

  The inputs to a layer are tensors, packaged according to how many there are:

    - `n_in = 0`: an empty tuple
    - `n_in = 1`: one tensor (NOT wrapped in a tuple)
    - `n_in > 1`: a tuple of tensors

  (The special treatment of the single-input case is meant to simplify the
  work of layer writers; this design choice may be revisited in the future.)

  The outputs from a layer are also tensors, packaged the same as layer inputs:

    - `n_out = 0`: an empty tuple
    - `n_out = 1`: the tensor (NOT wrapped in a tuple)
    - `n_out > 1`: a tuple of tensors

  The Trax runtime maintains a data stack with which layer calls are composed.
  For more complex data network architectures, possibly involving multiple data
  flows, one can view each layer as a function from stack state to stack state,
  where the function's inputs are a slice from the stack, and the function's
  outputs are spliced back into the stack.
  """

  def __init__(self, n_in=1, n_out=1, name=None, sublayers_to_print=None):
    """Creates a partially initialized, unconnected layer instance.

    Args:
      n_in: Number of inputs expected by this layer.
      n_out: Number of outputs promised by this layer.
      name: Class-like name for this layer; for use when printing this layer.
      sublayers_to_print: Sublayers to display when printing out this layer;
        if None (the default), display all sublayers.
    """
    self._n_in = n_in
    self._n_out = n_out
    self._name = self.__class__.__name__ if name is None else name
    self._sublayers_to_print = sublayers_to_print
    self._sublayers = ()  # Default is no sublayers.

    # The actual rng value/shape depends on the backend, which may not yet be
    # initialized at the point this method is run. Hence, at first initialize
    # only a seed random integer, in a backend-neutral way.
    self._rng = None
    self._rng_seed_int = random.randint(0, 2**31 - 1)

    # The private fields _weights and _state store the private part of
    # layer weights and state. When a layer has no sublayers, these are
    # the same as layer.weights and layer.state. For layers with sublayers
    # (i.e., combinators), these just mark which weights are cached -- see
    # the getter and setter for weights and state for details.
    # There is no need to use these fields in most user-implemented classes.
    self._weights = EMPTY_WEIGHTS  # By default no trainable weights.
    self._state = EMPTY_STATE  # By default no non-trainable state.

    # Record layer creation site for use in LayerError messages.
    # The frame can mutate, so copy relevant values out of it.
    frame = _find_frame(inspect.currentframe())
    self._caller = {'filename': copy.copy(frame.f_code.co_filename),
                    'lineno': int(frame.f_lineno)}
    del frame  # Just in case.

    self._init_cached = False
    self._jit_cache = {}

  def __repr__(self):
    """Renders this layer as a medium-detailed string, to help in debugging.

    Subclasses should aim for high-signal/low-noise when overriding this
    method.

    Returns:
      A high signal-to-noise string representing this layer.
    """
    def indent_string(x):
      return '  ' + x.replace('\n', '\n  ')

    name_str = self._name
    n_in, n_out = self.n_in, self.n_out
    if n_in != 1: name_str += f'_in{n_in}'
    if n_out != 1: name_str += f'_out{n_out}'

    if self._sublayers_to_print is not None:
      substructure = self._sublayers_to_print
    else:
      substructure = self.sublayers
    if substructure:
      substructure_strs = [str(x) for x in substructure if str(x)]
      substructure_str = '\n'.join(indent_string(s) for s in substructure_strs)
      return f'{name_str}[\n{substructure_str}\n]'
    else:
      return name_str

  def __call__(self, x, weights=None, state=None, rng=None):
    """Makes layers callable; for use in tests or interactive settings.

    This convenience method helps library users play with, test, or otherwise
    probe the behavior of layers outside of a full training environment. It
    presents the layer as callable function from inputs to outputs, with the
    option of manually specifying weights and non-parameter state per individual
    call. For convenience, weights and non-parameter state are cached per layer
    instance, starting from default values of `EMPTY_WEIGHTS` and `EMPTY_STATE`,
    and acquiring non-empty values either by initialization or from values
    explicitly provided via the weights and state keyword arguments, in which
    case the old weights will be preserved, and the state will be updated.

    Args:
      x: Zero or more input tensors, packaged as described in the `Layer` class
          docstring.
      weights: Weights or `None`; if `None`, use self's cached weights value.
      state: State or `None`; if `None`, use self's cached state value.
      rng: Single-use random number generator (JAX PRNG key), or `None`;
          if `None`, use a default computed from an integer 0 seed.

    Returns:
      Zero or more output tensors, packaged as described in the `Layer` class
      docstring.
    """
    weights = self.weights if weights is None else weights
    rng = self.rng if rng is None else rng
    if state is not None:
      self.state = state  # Needed if the model wasn't fully initialized.
    state = self.state
    outputs, new_state = self.pure_fn(x, weights, state, rng)
    self.state = new_state
    return outputs

  def forward(self, inputs):
    """Computes this layer's output as part of a forward pass through the model.

    A layer subclass overrides this method to define how the layer computes
    outputs from inputs. If the layer depends on weights, state, or randomness
    as part of the computation, the needed information can be accessed as
    properties of the layer object: `self.weights`, `self.state`, and
    `self.rng`. (See numerous examples in `trax.layers.core`.)

    Args:
      inputs: Zero or more input tensors, packaged as described in the `Layer`
          class docstring.

    Returns:
      Zero or more output tensors, packaged as described in the `Layer` class
      docstring.
    """
    raise NotImplementedError

  def init_weights_and_state(self, input_signature):
    """Initializes weights and state, to handle input with the given signature.

    A layer subclass must override this method if the layer uses weights or
    state. To initialize weights, set `self.weights` to desired (typically
    random) values. To initialize state (uncommon), set `self.state` to desired
    starting values.

    Args:
      input_signature: A `ShapeDtype` instance (if this layer takes one input)
          or a list/tuple of `ShapeDtype` instances.
    """
    del input_signature

  @property
  def has_backward(self):
    """Returns `True` if this layer provides its own custom backward pass code.

    A layer subclass that provides custom backward pass code (for custom
    gradients) must override this method to return `True`.
    """
    return False

  def backward(self, inputs, output, grad, weights, state, new_state, rng):
    """Custom backward pass to propagate gradients in a custom way.

    Args:
      inputs: Input tensors; can be a (possibly nested) tuple.
      output: The result of running this layer on inputs.
      grad: Gradient signal computed based on subsequent layers; its structure
          and shape must match output.
      weights: This layer's weights.
      state: This layer's state prior to the current forward pass.
      new_state: This layer's state after the current forward pass.
      rng: Single-use random number generator (JAX PRNG key).

    Returns:
      The custom gradient signal for the input. Note that we need to return
      a gradient for each argument of forward, so it will usually be a tuple
      of signals: the gradient for inputs and weights.
    """
    raise NotImplementedError

  # End of public subclassing interface.
  # Begin public callable interface.

  def init(self, input_signature, rng=None, use_cache=False):
    """Initializes weights/state of this layer and its sublayers recursively.

    Initialization creates layer weights and state, for layers that use them.
    It derives the necessary array shapes and data types from the layer's input
    signature, which is itself just shape and data type information.

    For layers without weights or state, this method safely does nothing.

    This method is designed to create weights/state only once for each layer
    instance, even if the same layer instance occurs in multiple places in the
    network. This enables weight sharing to be implemented as layer sharing.

    Args:
      input_signature: `ShapeDtype` instance (if this layer takes one input)
          or list/tuple of `ShapeDtype` instances.
      rng: Single-use random number generator (JAX PRNG key), or `None`;
          if `None`, use a default computed from an integer 0 seed.
      use_cache: If `True`, and if this layer instance has already been
          initialized elsewhere in the network, then return special marker
          values -- tuple `(GET_WEIGHTS_FROM_CACHE, GET_STATE_FROM_CACHE)`.
          Else return this layer's newly initialized weights and state.

    Returns:
      A `(weights, state)` tuple.
    """
    try:
      if self._init_cached and use_cache:
        return (GET_WEIGHTS_FROM_CACHE, GET_STATE_FROM_CACHE)

      if rng is not None:
        self.rng = rng
      self.init_weights_and_state(input_signature)

      if use_cache:
        self._init_cached = True
      else:
        self._clear_init_cache()

      return (self.weights, self.state)

    except Exception:
      # Skipping 3 lines as it's always the uninteresting internal call.
      name, trace = self._name, _short_traceback(skip=3)
      raise LayerError(name, 'init', self._caller,
                       input_signature, trace) from None

  def init_from_file(self, file_name, weights_only=False, input_signature=None):
    """Initializes this layer and its sublayers from a pickled checkpoint.

    In the common case (`weights_only=False`), the file must be a gziped pickled
    dictionary containing items with keys `'flat_weights', `'flat_state'` and
    `'input_signature'`, which are used to initialize this layer.
    If `input_signature` is specified, it's used instead of the one in the file.
    If `weights_only` is `True`, the dictionary does not need to have the
    `'flat_state'` item and the state it not restored either.

    Args:
      file_name: Name/path of the pickled weights/state file.
      weights_only: If `True`, initialize only the layer's weights. Else
          initialize both weights and state.
      input_signature: Input signature to be used instead of the one from file.

    Returns:
      A `(weights, state)` tuple.
    """
    with tf.io.gfile.GFile(file_name, 'rb') as f:
      with gzip.GzipFile(fileobj=f, compresslevel=2) as gzipf:
        dictionary = pickle.load(gzipf)
    # In the current checkpoint format, we store weights in a separate
    # non-pickled file with the same name but added ".npy".
    if isinstance(dictionary['flat_weights'], int):
      if file_name.endswith('.pkl.gz'):
        weights_path = file_name[:-6] + 'weights.npy.gz'
      else:
        weights_path = file_name + '.npy'
      if not tf.io.gfile.exists(weights_path):  # old format compatibility
        weights_path = file_name + '.npy'
      dictionary['flat_weights'] = np_from_file(
          weights_path, compresslevel=dictionary['flat_weights'])
    if input_signature is None:
      input_signature = dictionary['input_signature']
    if weights_only and input_signature is not None:
      self.init(input_signature)
    weights_and_state_sig = self.weights_and_state_signature(input_signature)
    weights, state = unflatten_weights_and_state(
        dictionary['flat_weights'], dictionary['flat_state'],
        weights_and_state_sig, weights_only=weights_only)
    if not weights_only:
      self.state = state
    self.weights = weights
    return (self.weights, self.state)

  def save_to_file(self, file_name, weights_only=False, input_signature=None):
    """Saves this layer and its sublayers to a pickled checkpoint.

    Args:
      file_name: Name/path of the pickled weights/state file.
      weights_only: If `True`, save only the layer's weights. Else
          save both weights and state.
      input_signature: Input signature to be used.
    """
    flat_weights, flat_state = flatten_weights_and_state(
        self.weights, self.state)
    dictionary = {
        'flat_weights': flat_weights,
    }
    if not weights_only:
      dictionary['flat_state'] = flat_state
    if input_signature is not None:
      dictionary['input_signature'] = input_signature

    tmp_file_path = file_name + '._tmp_'
    with tf.io.gfile.GFile(tmp_file_path, 'wb') as f:
      with gzip.GzipFile(fileobj=f, compresslevel=2) as gzipf:
        pickle.dump(dictionary, gzipf, protocol=pickle.HIGHEST_PROTOCOL)
    # Moving a file is much less error-prone than pickling large files.
    tf.io.gfile.rename(tmp_file_path, file_name, overwrite=True)

  # End of public callable methods.
  # Methods and properties below are reserved for internal use.

  @property
  def name(self):
    """Returns the name of this layer."""
    return self._name

  @property
  def n_in(self):
    """Returns how many tensors this layer expects as input."""
    return self._n_in

  @property
  def n_out(self):
    """Returns how many tensors this layer promises as output."""
    return self._n_out

  @property
  def sublayers(self):
    """Returns a tuple containing this layer's sublayers; may be empty."""
    return self._sublayers

  @property
  def weights(self):
    """Returns this layer's weights.

    Depending on the layer, the weights can be in the form of:

      - an empty tuple
      - a tensor (ndarray)
      - a nested structure of tuples and tensors

    If the layer has sublayers, the weights by convention will be
    a tuple of length `len(sublayers)` containing the weights of sublayers.
    Note that in this case self._weights only marks which ones are shared.
    """
    if not self.sublayers:
      return self._weights
    else:
      return tuple(layer.weights if w is None else w
                   for (layer, w) in zip(self.sublayers, self._weights))

  @weights.setter
  def weights(self, weights):
    """Sets the weights of this layer and its sublayers.

    Args:
      weights: the weights to set; if layer has sublayers, weights should be
        either a list or a tuple of the same length as `len(self.sublayers)`
        and it will be used to set the weights of all sublayers.
    """
    if isinstance(weights, dict) and weights == GET_WEIGHTS_FROM_CACHE:
      return
    if not self.sublayers:
      self._weights = weights
    else:
      # When having sublayers, self._weights just marks which are cached,
      # the actual weights are stored by sublayers.
      self._weights = []
      for w in weights:
        if isinstance(w, dict) and w == GET_WEIGHTS_FROM_CACHE:
          self._weights.append(w)
        else:
          self._weights.append(None)
      # Set sublayer weights.
      n_layers = len(self.sublayers)
      if len(weights) != n_layers:
        raise ValueError(
            f'Number of weight elements ({len(weights)}) does not equal the '
            f'number of sublayers ({n_layers}) in: {str(self)}.')
      for sublayer, sublayer_weights in zip(self.sublayers, weights):
        sublayer.weights = sublayer_weights

  @property
  def state(self):
    """Returns a tuple containing this layer's state; may be empty.

    If the layer has sublayers, the state by convention will be
    a tuple of length `len(sublayers)` containing sublayer states.
    Note that in this case self._state only marks which ones are shared.
    """
    if not self.sublayers:
      return self._state
    else:
      return tuple(layer.state if s is None else s
                   for (layer, s) in zip(self.sublayers, self._state))

  @state.setter
  def state(self, state):
    """Sets the state of this layer and its sublayers.

    Args:
      state: the state to set; if layer has sublayers, state should be
        either a list or a tuple of the same length as `len(self.sublayers)`
        and it will be used to set the state of all sublayers.
    """
    if isinstance(state, dict) and state == GET_STATE_FROM_CACHE:
      return
    if not self._sublayers:
      self._state = state
    else:
      # When having sublayers, self._state just marks which are cached,
      # the actual weights are stored by sublayers.
      self._state = []
      for s in state:
        if isinstance(s, dict) and s == GET_STATE_FROM_CACHE:
          self._state.append(s)
        else:
          self._state.append(None)
      # Set sublayer states.
      n_layers = len(self.sublayers)
      if len(state) != n_layers:
        raise ValueError(
            f'Number of state elements ({len(state)}) does not equal the '
            f'number of sublayers ({n_layers}) in: {str(self)}.')
      for sublayer, sublayer_state in zip(self.sublayers, state):
        sublayer.state = sublayer_state

  def weights_and_state_signature(self, input_signature, unsafe=False):
    """Return a pair containing the signatures of weights and state."""
    rng, state, weights = self.rng, self.state, self.weights
    abstract_init = fastmath.abstract_eval(self.init)
    sig = abstract_init(input_signature)
    self.rng = rng
    if not unsafe:
      self.state, self.weights = state, weights
    return sig

  @property
  def rng(self):
    """Returns this layer's current single-use random number generator.

    Code that wants to base random samples on this generator must explicitly
    split off new generators from it. (See, for example, the `rng` setter code
    below.)
    """
    if self._rng is None:
      # One-time initialization from backend-neutral seed int.
      self._rng = fastmath.random.get_prng(self._rng_seed_int)
    return self._rng

  @rng.setter
  def rng(self, rng):
    """Sets the rng (JAX PRNG key) for this layer and sublayers, recursively."""
    self._rng = rng
    sublayers = self.sublayers
    if sublayers:
      rngs = fastmath.random.split(rng, len(sublayers))
      for sublayer, rng in zip(sublayers, rngs):
        sublayer.rng = rng

  def _clear_init_cache(self):
    self._init_cached = False
    for sublayer in self.sublayers:
      sublayer._clear_init_cache()  # pylint: disable=protected-access

  def pure_fn(self, x, weights, state, rng, use_cache=False):
    """Applies this layer as a pure function with no optional args.

    This method exposes the layer's computation as a pure function. This is
    especially useful for JIT compilation. Do not override, use `forward`
    instead.

    Args:
      x: Zero or more input tensors, packaged as described in the `Layer` class
          docstring.
      weights: A tuple or list of trainable weights, with one element for this
          layer if this layer has no sublayers, or one for each sublayer if
          this layer has sublayers. If a layer (or sublayer) has no trainable
          weights, the corresponding weights element is an empty tuple.
      state: Layer-specific non-parameter state that can update between batches.
      rng: Single-use random number generator (JAX PRNG key).
      use_cache: if `True`, cache weights and state in the layer object; used
        to implement layer sharing in combinators.

    Returns:
      A tuple of `(tensors, state)`. The tensors match the number (`n_out`)
      promised by this layer, and are packaged as described in the `Layer`
      class docstring.
    """
    try:
      old_weights, old_state, old_rng = self.weights, self.state, self.rng
      self._rng = rng
      # The isinstance check is only needed when == is overloaded, as in TF.
      if (isinstance(weights, dict) and isinstance(state, dict) and
          weights == GET_WEIGHTS_FROM_CACHE and state == GET_STATE_FROM_CACHE):
        was_cached = True
        weights = self.weights
        state = self.state
      else:
        # In this case, we're called for the first time: cache weights.
        was_cached = False
        self.weights, self.state = weights, state

      # If weights are sharded across multiple devices, unshard before forward.
      sharded_weights, weights_were_unsharded = weights, False
      if N_WEIGHTS_SHARDS > 1 and not self.sublayers:
        self.weights, weights_were_unsharded = unshard_in_pmap(
            weights, N_WEIGHTS_SHARDS)

      if not self.has_backward:
        outputs = self.forward(x)
        s = self.state
      else:
        outputs, s = self._do_custom_gradients(x)
        self.state = s
      self._rng = old_rng
      if weights_were_unsharded:  # only store a shard of weights if sharded
        self.weights = sharded_weights

      if not use_cache:
        self.weights, self.state = old_weights, old_state
      if was_cached:  # If the layer was shared, return a state marking this.
        s = GET_STATE_FROM_CACHE
      return outputs, s

    except Exception:
      # Skipping 3 lines as it's always the uninteresting internal call.
      name, trace = self._name, _short_traceback(skip=3)
      raise LayerError(name, 'pure_fn',
                       self._caller, signature(x), trace) from None

  def output_signature(self, input_signature):
    """Returns output signature this layer would give for `input_signature`."""
    return self._forward_abstract(input_signature)[0]  # output only, not state

  def _forward_abstract(self, input_signature):
    """Computes shapes and dtypes this layer would produce in a forward pass.

    Args:
      input_signature: `ShapeDtype` instance (if this layer takes one input)
          or list/tuple of `ShapeDtype` instances.

    Returns:
      Tuple of (output, state).

      The output part of the tuple is a `ShapeDtype` instance representing the
      shape and type of the output (if this layer has one output) or a tuple
      of `ShapeDtype` instances (if this layer has more than one output).
    """
    try:
      # Note: By using rng_signature in place of an rng, we avoid computing and
      # permanently storing in global memory a large number of dropout masks.
      # TODO(jonni): Check if using an rng still carries this cost.
      dummy_rng = fastmath.random.get_prng(0)
      rng_signature = ShapeDtype(dummy_rng.shape, dummy_rng.dtype)
      weights_signature = nested_map(signature, self.weights)
      state_signature = nested_map(signature, self.state)
      forward_infer_shapes = fastmath.abstract_eval(self.pure_fn)
      return forward_infer_shapes(
          input_signature, weights_signature, state_signature, rng_signature)
    except Exception:
      # TODO(lukaszkaiser): the choice of 7 is a heuristic, can we automate it?
      # Skipping 7 lines which are all JAX abstract'ifying wrappers.
      name, trace = self._name, _short_traceback(skip=7)
      raise LayerError(name, '_forward_abstract', self._caller, input_signature,
                       trace) from None

  # pylint: disable=protected-access
  def _do_custom_gradients(self, x):
    """Calls this layer for a forward pass, but with custom gradients."""

    def _f(state, rng, y, weights):
      old_weights, old_state, old_rng = self.weights, self.state, self._rng
      self.weights, self.state, self._rng = weights, state, rng
      res = self.forward(y)
      s = self.state
      self.weights, self.state, self._rng = old_weights, old_state, old_rng
      return res, s

    def _f_fwd(state, rng, y, weights):
      old_weights, old_state, old_rng = self.weights, self.state, self._rng
      self.weights, self.state, self._rng = weights, state, rng
      res = self.forward(y)
      s = self.state
      self.weights, self.state, self._rng = old_weights, old_state, old_rng
      return (res, s), (state, rng, y, res, weights, s)

    def _f_bwd(residual, grad):
      """Custom gradient function."""
      state, rng, y, output, weights, new_state = residual
      grad = grad[0]  # Ignore dummy gradient wrt state.
      out = self.backward(y, output, grad, weights, state, new_state, rng)
      return (None, None, *out)

    do_forward = fastmath.custom_vjp(_f, _f_fwd, _f_bwd, nondiff_argnums=(0, 1))

    output, state = do_forward(self.state, self._rng, x, self.weights)
    return output, state

  def _settable_attrs(self):
    """We only allow to set these attributes in Trax layers to prevent typos."""
    return ('weights', 'state', 'rng')

  def __setattr__(self, attr, value):
    """Sets class attributes and protects from typos.

    In Trax layers, we only allow to set the following public attributes::

      - weights
      - state
      - rng

    This function prevents from setting other public attributes to avoid typos,
    for example, this is not possible and would be without this function::

      [typo]   layer.weighs = some_tensor

    If you need to set other public attributes in a derived class (which we
    do not recommend as in almost all cases it suffices to use a private
    attribute), override self._settable_attrs to include the attribute name.

    Args:
      attr: Name of the attribute to be set.
      value: Value to be assigned to the attribute.
    """
    if attr[0] != '_' and attr not in self._settable_attrs():
      raise ValueError(
          f'Trax layers only allow to set {self._settable_attrs()} as public '
          f'attribues, not {attr}.')
    else:
      super().__setattr__(attr, value)


class PureLayer(Layer):
  """Pure function from inputs to outputs, packaged as neural network layer.

  The `PureLayer` class represents the simplest kinds of layers: layers with
  no trainable weights and no randomness, hence pure functions from inputs to
  outputs.
  """

  def __init__(self, forward_fn, n_in=1, n_out=1, name='PureLayer'):
    """Creates an unconnected `PureLayer` instance.

    Args:
      forward_fn: Pure function from input tensors to output tensors, where
          inputs and outputs are packaged as specified for `forward`.
      n_in: Number of inputs expected by this layer.
      n_out: Number of outputs promised by this layer.
      name: Class-like name for this layer; for use only in debugging.
    """
    super().__init__(n_in, n_out, name)
    self._forward_fn = forward_fn

  def forward(self, inputs):
    """Overrides `Layer.forward`.

    Args:
      inputs: Zero or more input tensors, packaged as described in the `Layer`
          class docstring.

    Returns:
      Zero or more output tensors, packaged as described in the `Layer` class
      docstring.
    """
    _validate_forward_input(inputs, self.n_in)
    raw_output = self._forward_fn(inputs)
    output = () if _is_empty(raw_output) else raw_output
    return output


def Fn(name, f, n_out=1):  # pylint: disable=invalid-name
  """Returns a layer with no weights that applies the function `f`.

  `f` can take and return any number of arguments, and takes only positional
  arguments -- no default or keyword arguments. It often uses JAX-numpy (`jnp`).
  The following, for example, would create a layer that takes two inputs and
  returns two outputs -- element-wise sums and maxima:

      `Fn('SumAndMax', lambda x0, x1: (x0 + x1, jnp.maximum(x0, x1)), n_out=2)`

  The layer's number of inputs (`n_in`) is automatically set to number of
  positional arguments in `f`, but you must explicitly set the number of
  outputs (`n_out`) whenever it's not the default value 1.

  Args:
    name: Class-like name for the resulting layer; for use in debugging.
    f: Pure function from input tensors to output tensors, where each input
        tensor is a separate positional arg, e.g., `f(x0, x1) --> x0 + x1`.
        Output tensors must be packaged as specified in the `Layer` class
        docstring.
    n_out: Number of outputs promised by the layer; default value 1.

  Returns:
    Layer executing the function `f`.
  """
  argspec = inspect.getfullargspec(f)
  if argspec.defaults is not None:
    raise ValueError('Function has default arguments (not allowed).')
  if argspec.varkw is not None:
    raise ValueError('Function has keyword arguments (not allowed).')
  if argspec.varargs is not None:
    raise ValueError('Function has variable args (not allowed).')

  def _forward(xs):  # pylint: disable=invalid-name
    if not isinstance(xs, (tuple, list)):
      xs = (xs,)
    return f(*xs)

  n_in = len(argspec.args)
  name = name or 'Fn'
  return PureLayer(_forward, n_in=n_in, n_out=n_out, name=name)


class LayerError(Exception):
  """Exception raised in the layer stack."""

  def __init__(self, layer_name, function_name, caller,
               input_signature, traceback_string):
    self._layer_name = layer_name
    self._function_name = function_name
    self._caller = caller  # Python inspect object with init caller info.
    self._traceback = traceback_string
    self._input_signature = input_signature
    super().__init__(self.message)

  @property
  def message(self):
    """Assembles current layer context into an error message."""
    prefix = 'Exception passing through layer '
    prefix += '%s (in %s):\n' % (self._layer_name, self._function_name)
    short_path = '[...]/' + '/'.join(
        self._caller['filename'].split('/')[-3:])
    caller = '  layer created in file %s, line %d\n' % (short_path,
                                                        self._caller['lineno'])
    shapes_str = '  layer input shapes: %s\n\n' % str(self._input_signature)
    return prefix + caller + shapes_str + self._traceback


def flatten_weights_and_state(weights, state):
  """Flatten weights and state into lists, excluding empty and cached ones."""
  def _is_empty_weight(x):
    return (x is EMPTY_WEIGHTS or
            (isinstance(x, dict) and x == GET_WEIGHTS_FROM_CACHE))
  flat_weights = [w for w in fastmath.tree_flatten(weights)
                  if not _is_empty_weight(w)]
  def _is_empty_state(x):
    return (x is EMPTY_STATE or
            (isinstance(x, dict) and x == GET_STATE_FROM_CACHE))
  flat_state = [s for s in fastmath.tree_flatten(state)
                if not _is_empty_state(s)]
  return flat_weights, flat_state


def unflatten_weights_and_state(
    flat_weights, flat_state, weights_and_state_signature, weights_only=False):
  """Unflatten weights and state given their signatures."""
  weights_tree, state_tree = weights_and_state_signature
  weights_to_copy = [EMPTY_WEIGHTS, GET_WEIGHTS_FROM_CACHE]
  weights, _ = fastmath.tree_unflatten(flat_weights, weights_tree,
                                       copy_from_tree=weights_to_copy)
  state = None
  if not weights_only:
    states_to_copy = [EMPTY_STATE, GET_STATE_FROM_CACHE]
    state, _ = fastmath.tree_unflatten(flat_state, state_tree,
                                       copy_from_tree=states_to_copy)
  return weights, state


def np_to_file(list_of_nparrays, file_path, compresslevel):
  """Save numpy arrays to file_path with gzipping and failure protection."""
  # Pickle to tmp file and overwrite to prevent writing partial files.
  tmp_file_path = file_path + '._tmp_'
  with tf.io.gfile.GFile(tmp_file_path, 'wb') as f:
    with gzip.GzipFile(fileobj=f, compresslevel=compresslevel) as gzipf:
      for x in list_of_nparrays:
        np.save(gzipf, x, allow_pickle=False)
  # Moving a file is much less error-prone than pickling large files.
  tf.io.gfile.rename(tmp_file_path, file_path, overwrite=True)


def np_from_file(file_path, compresslevel):
  """Load numpy arrays from file_path with gzipping."""
  if not tf.io.gfile.exists(file_path):
    raise FileNotFoundError(file_path)
  res = []
  with tf.io.gfile.GFile(file_path, 'rb') as f:
    with gzip.GzipFile(fileobj=f, compresslevel=compresslevel) as gzipf:
      while True:
        try:
          res.append(np.load(gzipf, allow_pickle=False))
        except Exception:  # pylint: disable=broad-except
          break
  return res


def to_list(outputs):
  """Converts layer outputs to a nested list, for easier equality testing.

  Args:
    outputs: A tensor or tuple/list of tensors coming from the forward
        application of a layer. Each tensor is NumPy ndarray-like, which
        complicates simple equality testing (e.g., via `assertEquals`):
        such tensors require equality testing to use either `all` (all
        elements match) or `any` (at least one element matches), which is not
        directly supported in `absltest`.

  Returns:
    A nested list structure containing all the output values, but now directly
    testable using `assertEquals`.
  """
  if isinstance(outputs, (list, tuple)):
    return [y.tolist() for y in outputs]
  else:
    return outputs.tolist()


def _validate_forward_input(x, n_in):
  if n_in != 1:
    if not isinstance(x, (tuple, list)):
      raise TypeError(
          f'Expected input to be a tuple or list; instead got {type(x)}.')
    if len(x) != n_in:
      raise ValueError(f'Input tuple length ({len(x)}) does not equal required '
                       f'number of inputs ({n_in}).')


def _is_empty(container):
  if container is None:
    raise ValueError('Argument "container" is None.')
  return isinstance(container, (list, tuple)) and len(container) == 0  # pylint: disable=g-explicit-length-test


def _find_frame(frame):
  """Find the frame with the caller on the stack."""
  def _dirname_is_trax_layers_or_gin(frame):
    """Skip frames coming from trax/layers or .../gin."""
    try:
      dirname1 = frame.f_code.co_filename.split('/')[-3]
      dirname2 = frame.f_code.co_filename.split('/')[-2]
      return (dirname1 == 'trax' and dirname2 == 'layers') or dirname2 == 'gin'
    except IndexError:
      return False

  while _dirname_is_trax_layers_or_gin(frame):
    frame = frame.f_back
  return frame


def _shorten_file_path(line):
  """Shorten file path in error lines for more readable tracebacks."""
  start = line.lower().find('file')
  if start < 0:
    return line
  first_quote = line.find('"', start)
  if first_quote < 0:
    return line
  second_quote = line.find('"', first_quote + 1)
  if second_quote < 0:
    return line
  path = line[first_quote + 1:second_quote]
  new_path = '/'.join(path.split('/')[-3:])
  return line[:first_quote] + '[...]/' + new_path + line[second_quote + 1:]


def _short_traceback(skip=3):
  """Cleaned-up form of traceback."""
  counter, res = 0, []
  # Skipping 3 lines by default: the top (useless) and self-call.
  # In python 3, we need to set chain to False (it doesn't exist in python 2).
  lines = traceback.format_exc(chain=False).splitlines()[skip:]  # pylint: disable=unexpected-keyword-arg
  for l in lines:
    if l.startswith('trax.layers.base.LayerError'):
      l = l[len('trax.layers.base.'):]  # Remove the trax.layers.base prefix.
    res.append(_shorten_file_path(l))
    if counter % 2 == 1:
      res.append('')
    counter += 1
    # If we see a LayerError, the traceback has already been processed.
    if l.startswith('LayerError'):
      # Skip 4 back except last as these are internal base-layer calls.
      res = res[:-4] + [res[-1]]
      res += lines[counter:]
      break
  return '\n'.join(res)


def _random_values(input_signature, rng):
  """Creates random floats or ints of the given shape.

  Args:
    input_signature: A `ShapeDtype` instance (if `layer_obj` takes one input)
        or a list/tuple of ShapeDtype instances.
    rng: Single-use random number generator (JAX PRNG key).

  Returns:
    Random values with the shape and type specified.
  """
  if isinstance(input_signature, ShapeDtype):
    shape, dtype = input_signature.shape, input_signature.dtype
    if np.issubdtype(dtype, np.integer):
      return fastmath.random.bernoulli(rng, 0.5, shape).astype(np.int32)
    else:
      return fastmath.random.uniform(rng, shape, minval=-1.0, maxval=1.0)
  elif isinstance(input_signature, (list, tuple)):
    return tuple(_random_values(x, rng) for x in input_signature)
  else:
    raise TypeError(type(input_signature))


def _shapes(x):
  """Gets a structure of shapes for a structure of nested arrays."""
  def shape(x):
    try:
      return tuple([int(i) for i in x.shape])
    except Exception:  # pylint: disable=broad-except
      return ()
  return tuple(nested_map(shape, x))


@functools.partial(fastmath.pmap, axis_name='batch')
def _axis_index(unused_x):
  """Return the axis indices."""
  return jax.lax.axis_index('batch')


def _axis_to_shard_heuristic(shape):
  """Chooses an axis to shard on - a simple heuristic to be revisited."""
  axis = 0 if len(shape) < 3 else -1
  return axis


def shard(tensors, n_shards=None):
  """Shard tensors across n_shards."""
  n_shards = N_WEIGHTS_SHARDS if n_shards is None else n_shards
  indices = _axis_index(np.zeros(fastmath.local_device_count()))
  def _shard_fn(x):
    axis = _axis_to_shard_heuristic(x.shape)
    if int(x.shape[axis]) % n_shards != 0:
      raise ValueError(f'Cannot split x with shape {x.shape} into {n_shards}.')
    split_x = jnp.split(x, n_shards, axis=axis)
    split_x = [split_x[i % n_shards] for i in indices]
    return np.stack(split_x, axis=0)
  return fastmath.nested_map(_shard_fn, tensors)


def unshard_in_pmap(tensors, n_shards):
  """Unshard tensors that were sharded into n_shards (call inside pmap)."""
  groups = [[n_shards * i + d for d in range(n_shards)]
            for i in range(fastmath.global_device_count() // n_shards)]
  def _unshard_fn(x):
    y = jax.lax.all_gather(x, 'batch', axis_index_groups=groups)
    split_y = jnp.split(y, n_shards, axis=0)
    split_y = [jnp.squeeze(sy, axis=0) for sy in split_y]
    axis = _axis_to_shard_heuristic(split_y[0].shape)
    return jnp.concatenate(split_y, axis=axis)
  try:
    jax.lax.axis_index('batch')  # will throw if not in pmap, e.g., on init
    res = fastmath.nested_map(_unshard_fn, tensors)
    return res, True
  except NameError:  # thrown from axis_index above
    return tensors, False


@functools.partial(fastmath.pmap, axis_name='batch')
def _all_gather(x, groups):
  return jax.lax.all_gather(x, 'batch', axis_index_groups=groups)


def unshard(tensors, n_shards=None):
  """Unshard tensors that were sharded into n_shards (outside of pmap)."""
  n_shards = N_WEIGHTS_SHARDS if n_shards is None else n_shards
  def _unshard_fn(x):
    # We use numpy here to put the large un-sharded arrays in CPU memory.
    # For unsharding on accelerators use ushard_in_pmap above and pmap it.
    split_y = np.split(np.asarray(x), n_shards, axis=0)
    split_y = [np.squeeze(sy, axis=0) for sy in split_y]
    axis = _axis_to_shard_heuristic(split_y[0].shape)
    return np.concatenate(split_y, axis=axis)
  return fastmath.nested_map(_unshard_fn, tensors)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL training."""

import functools

from absl.testing import absltest

from trax import layers as tl
from trax import models
from trax import optimizers as opt
from trax import test_utils
from trax.rl import actor_critic_joint
from trax.rl import task as rl_task
from trax.supervised import lr_schedules



class ActorCriticJointTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')

  def test_awrjoint_save_restore(self):
    """Check save and restore of joint AWR trainer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=2,
                          max_steps=2)
    joint_model = functools.partial(
        models.PolicyAndValue,
        body=lambda mode: tl.Serial(tl.Dense(4), tl.Relu()),
    )
    tmp_dir = self.create_tempdir().full_path
    trainer1 = actor_critic_joint.AWRJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.Adam,
        batch_size=4,
        train_steps_per_epoch=1,
        n_trajectories_per_epoch=2,
        output_dir=tmp_dir)
    trainer1.run(2)
    self.assertEqual(trainer1.current_epoch, 2)
    self.assertEqual(trainer1._trainer.step, 2)
    # Agent 2 starts where agent 1 stopped.
    trainer2 = actor_critic_joint.AWRJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.Adam,
        batch_size=4,
        train_steps_per_epoch=1,
        n_trajectories_per_epoch=2,
        output_dir=tmp_dir)
    trainer2.run(1)
    self.assertEqual(trainer2.current_epoch, 3)
    self.assertEqual(trainer2._trainer.step, 3)
    trainer1.close()
    trainer2.close()


  def test_jointppotrainer_cartpole(self):
    """Test-runs joint PPO on CartPole."""

    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=2)
    joint_model = functools.partial(
        models.PolicyAndValue,
        body=lambda mode: tl.Serial(tl.Dense(2), tl.Relu()),
    )
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')

    trainer = actor_critic_joint.PPOJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.Adam,
        lr_schedule=lr,
        batch_size=4,
        train_steps_per_epoch=2,
        n_trajectories_per_epoch=5)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_jointawrtrainer_cartpole(self):
    """Test-runs joint AWR on cartpole."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=1,
                          max_steps=2)
    joint_model = functools.partial(
        models.PolicyAndValue,
        body=lambda mode: tl.Serial(tl.Dense(64), tl.Relu()),
    )
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic_joint.AWRJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.Adam,
        lr_schedule=lr,
        batch_size=4,
        train_steps_per_epoch=2,
        n_trajectories_per_epoch=5)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_jointa2ctrainer_cartpole(self):
    """Test-runs joint A2C on cartpole."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=1,
                          max_steps=2)
    joint_model = functools.partial(
        models.PolicyAndValue,
        body=lambda mode: tl.Serial(tl.Dense(64), tl.Relu()),
    )
    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')
    trainer = actor_critic_joint.A2CJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.RMSProp,
        lr_schedule=lr,
        batch_size=2,
        train_steps_per_epoch=1,
        n_trajectories_per_epoch=1)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_jointawrtrainer_cartpole_transformer(self):
    """Test-runs joint AWR on cartpole with Transformer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=1,
                          max_steps=2)
    body = lambda mode: models.TransformerDecoder(  # pylint: disable=g-long-lambda
        d_model=4, d_ff=4, n_layers=1, n_heads=1, mode=mode)
    joint_model = functools.partial(models.PolicyAndValue, body=body)
    trainer = actor_critic_joint.AWRJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.Adam,
        batch_size=4,
        train_steps_per_epoch=2,
        n_trajectories_per_epoch=2,
        max_slice_length=2)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)

  def test_jointa2ctrainer_cartpole_transformer(self):
    """Test-runs joint A2C on cartpole with Transformer."""
    task = rl_task.RLTask('CartPole-v0', initial_trajectories=1,
                          max_steps=2)
    body = lambda mode: models.TransformerDecoder(  # pylint: disable=g-long-lambda
        d_model=4, d_ff=4, n_layers=1, n_heads=1, mode=mode)
    joint_model = functools.partial(models.PolicyAndValue, body=body)
    trainer = actor_critic_joint.A2CJoint(
        task,
        joint_model=joint_model,
        optimizer=opt.RMSProp,
        batch_size=4,
        train_steps_per_epoch=2,
        n_trajectories_per_epoch=2)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)


if __name__ == '__main__':
  absltest.main()
